from unittest.mock import patch

import pytest
from django.contrib.auth.models import User
from django.core.management import call_command
from django.db.models import BigAutoField
from django.test import TestCase

from ninja_jwt.exceptions import TokenError
from ninja_jwt.schema import TokenVerifySerializer
from ninja_jwt.settings import api_settings
from ninja_jwt.token_blacklist.models import BlacklistedToken, OutstandingToken
from ninja_jwt.tokens import AccessToken, RefreshToken, SlidingToken
from ninja_jwt.utils import aware_utcnow, datetime_from_epoch

from .utils import MigrationTestCase, override_api_settings


@pytest.mark.django_db
class TestTokenBlacklist:
    @pytest.fixture(autouse=True)
    def setUp(self):
        self.user = User.objects.create(
            username="test_user",
            password="test_password",
        )

    def test_sliding_tokens_are_added_to_outstanding_list(self):
        token = SlidingToken.for_user(self.user)

        qs = OutstandingToken.objects.all()
        outstanding_token = qs.first()

        assert qs.count() == 1
        assert outstanding_token.user == self.user
        assert outstanding_token.jti == token["jti"]
        assert outstanding_token.token == str(token)
        assert outstanding_token.created_at == token.current_time
        assert outstanding_token.expires_at == datetime_from_epoch(token["exp"])

    def test_refresh_tokens_are_added_to_outstanding_list(self):
        token = RefreshToken.for_user(self.user)

        qs = OutstandingToken.objects.all()
        outstanding_token = qs.first()

        assert qs.count() == 1
        assert outstanding_token.user == self.user
        assert outstanding_token.jti == token["jti"]
        assert outstanding_token.token == str(token)
        assert outstanding_token.created_at == token.current_time
        assert outstanding_token.expires_at == datetime_from_epoch(token["exp"])

    def test_access_tokens_are_not_added_to_outstanding_list(self):
        AccessToken.for_user(self.user)

        qs = OutstandingToken.objects.all()

        assert not qs.exists()

    def test_token_will_not_validate_if_blacklisted(self):
        token = RefreshToken.for_user(self.user)
        outstanding_token = OutstandingToken.objects.first()

        # Should raise no exception
        RefreshToken(str(token))

        # Add token to blacklist
        BlacklistedToken.objects.create(token=outstanding_token)

        with pytest.raises(TokenError) as e:
            # Should raise exception
            RefreshToken(str(token))
            assert "blacklisted" in e.exception.args[0]

    def test_tokens_can_be_manually_blacklisted(self):
        token = RefreshToken.for_user(self.user)

        # Should raise no exception
        RefreshToken(str(token))

        assert OutstandingToken.objects.count() == 1

        # Add token to blacklist
        blacklisted_token, created = token.blacklist()

        # Should not add token to outstanding list if already present
        assert OutstandingToken.objects.count() == 1

        # Should return blacklist record and boolean to indicate creation
        assert blacklisted_token.token.jti == token["jti"]
        assert created

        with pytest.raises(TokenError) as e:
            # Should raise exception
            RefreshToken(str(token))
            assert "blacklisted" in e.exception.args[0]

        # If blacklisted token already exists, indicate no creation through
        # boolean
        blacklisted_token, created = token.blacklist()
        assert blacklisted_token.token.jti == token["jti"]
        assert not created

        # Should add token to outstanding list if not already present
        new_token = RefreshToken()
        blacklisted_token, created = new_token.blacklist()
        assert blacklisted_token.token.jti == new_token["jti"]
        assert created

        assert OutstandingToken.objects.count() == 2


@pytest.mark.django_db
class TestTokenBlacklistFlushExpiredTokens:
    @pytest.fixture(autouse=True)
    def setUp(self):
        self.user = User.objects.create(
            username="test_user",
            password="test_password",
        )

    def test_it_should_delete_any_expired_tokens(self):
        # Make some tokens that won't expire soon
        not_expired_1 = RefreshToken.for_user(self.user)
        not_expired_2 = RefreshToken.for_user(self.user)
        not_expired_3 = RefreshToken()

        # Blacklist fresh tokens
        not_expired_2.blacklist()
        not_expired_3.blacklist()

        # Make tokens with fake exp time that will expire soon
        fake_now = aware_utcnow() - api_settings.REFRESH_TOKEN_LIFETIME

        with patch("ninja_jwt.tokens.aware_utcnow") as fake_aware_utcnow:
            fake_aware_utcnow.return_value = fake_now
            expired_1 = RefreshToken.for_user(self.user)
            expired_2 = RefreshToken()

        # Blacklist expired tokens
        expired_1.blacklist()
        expired_2.blacklist()

        # Make another token that won't expire soon
        not_expired_4 = RefreshToken.for_user(self.user)

        # Should be certain number of outstanding tokens and blacklisted
        # tokens
        assert OutstandingToken.objects.count() == 6
        assert BlacklistedToken.objects.count() == 4

        call_command("flushexpiredtokens")

        # Expired outstanding *and* blacklisted tokens should be gone
        assert OutstandingToken.objects.count() == 4
        assert BlacklistedToken.objects.count() == 2

        assert [i.jti for i in OutstandingToken.objects.order_by("id")] == [
            not_expired_1["jti"],
            not_expired_2["jti"],
            not_expired_3["jti"],
            not_expired_4["jti"],
        ]

        assert [i.token.jti for i in BlacklistedToken.objects.order_by("id")] == [
            not_expired_2["jti"],
            not_expired_3["jti"],
        ]


@pytest.mark.django_db
class TestTokenVerifySerializerShouldHonourBlacklist(MigrationTestCase):
    migrate_from = ("token_blacklist", "0002_outstandingtoken_jti_hex")
    migrate_to = ("token_blacklist", "0003_auto_20171017_2007")

    @pytest.fixture(autouse=True)
    def setUp(self):
        self.user = User.objects.create(
            username="test_user",
            password="test_password",
        )
        super().setUp()

    def test_token_verify_serializer_should_honour_blacklist_if_blacklisting_enabled(
        self, monkeypatch
    ):
        with monkeypatch.context() as m:
            m.setattr(api_settings, "BLACKLIST_AFTER_ROTATION", True)
            refresh_token = RefreshToken.for_user(self.user)
            refresh_token.blacklist()

            with pytest.raises(Exception):
                TokenVerifySerializer(token=str(refresh_token))

    def test_token_verify_serializer_should_not_honour_blacklist_if_blacklisting_not_enabled(
        self, monkeypatch
    ):
        with monkeypatch.context() as m:
            m.setattr(api_settings, "BLACKLIST_AFTER_ROTATION", False)
            refresh_token = RefreshToken.for_user(self.user)
            refresh_token.blacklist()

            serializer = TokenVerifySerializer(token=str(refresh_token))
            assert serializer.token == str(refresh_token)


@pytest.mark.django_db
class TestBigAutoFieldIDMigration(MigrationTestCase):
    migrate_from = ("token_blacklist", "0007_auto_20171017_2214")
    migrate_to = ("token_blacklist", "0008_migrate_to_bigautofield")

    @pytest.fixture(autouse=True)
    def setUp(self):
        self.user = User.objects.create(
            username="test_user",
            password="test_password",
        )
        super().setUp()

    def test_outstandingtoken_id_field_is_biagauto_field(self):
        OutstandingToken = self.apps.get_model("token_blacklist", "OutstandingToken")
        assert isinstance(OutstandingToken._meta.get_field("id"), BigAutoField)

    def test_blacklistedtoken_id_field_is_biagauto_field(self):
        BlacklistedToken = self.apps.get_model("token_blacklist", "BlacklistedToken")
        assert isinstance(BlacklistedToken._meta.get_field("id"), BigAutoField)



from .backends import TokenBackend
from .settings import api_settings

token_backend = TokenBackend(
    api_settings.ALGORITHM,
    api_settings.SIGNING_KEY,
    api_settings.VERIFYING_KEY,
    api_settings.AUDIENCE,
    api_settings.ISSUER,
    api_settings.JWK_URL,
    api_settings.LEEWAY,
)



from datetime import timedelta
from importlib import reload
from unittest.mock import patch

import pytest
from django.contrib.auth import get_user_model
from django.utils import timezone

from ninja_jwt import schema
from ninja_jwt.settings import api_settings
from ninja_jwt.tokens import AccessToken, RefreshToken, SlidingToken
from ninja_jwt.utils import aware_utcnow, datetime_from_epoch, datetime_to_epoch

from .utils import APIViewTestCase, override_api_settings

User = get_user_model()


@pytest.mark.django_db
class TestTokenObtainPairView(APIViewTestCase):
    view_name = "jwt:token_obtain_pair"

    @pytest.fixture(autouse=True)
    def setUp(self):
        super().setUp()
        self.username = "test_user"
        self.password = "test_password"

        self.user = User.objects.create_user(
            username=self.username,
            password=self.password,
        )

    def test_fields_missing(self):
        res = self.view_post(data={}, content_type="application/json")
        assert res.status_code == 400
        assert User.USERNAME_FIELD in res.data
        assert "password" in res.data

        res = self.view_post(
            data={User.USERNAME_FIELD: self.username}, content_type="application/json"
        )
        assert res.status_code == 400
        assert "password" in res.data

        res = self.view_post(
            data={"password": self.password}, content_type="application/json"
        )
        assert res.status_code == 400
        assert User.USERNAME_FIELD in res.data

    def test_credentials_wrong(self):
        res = self.view_post(
            data={
                User.USERNAME_FIELD: self.username,
                "password": "test_user",
            },
            content_type="application/json",
        )
        assert res.status_code == 401
        assert "detail" in res.data

    def test_user_inactive(self):
        self.user.is_active = False
        self.user.save()

        res = self.view_post(
            data={
                User.USERNAME_FIELD: self.username,
                "password": self.password,
            },
            content_type="application/json",
        )
        assert res.status_code == 401
        assert "detail" in res.data

    def test_success(self):
        res = self.view_post(
            data={
                User.USERNAME_FIELD: self.username,
                "password": self.password,
            },
            content_type="application/json",
        )
        assert res.status_code == 200
        assert "access" in res.data
        assert "refresh" in res.data

    def test_update_last_login(self, monkeypatch):
        self.view_post(
            data={
                User.USERNAME_FIELD: self.username,
                "password": self.password,
            },
            content_type="application/json",
        )

        # verify last_login is not updated
        user = User.objects.get(username=self.username)
        assert user.last_login is None

        # verify last_login is updated
        with monkeypatch.context() as m:
            m.setattr(api_settings, "UPDATE_LAST_LOGIN", True)
            self.view_post(
                data={
                    User.USERNAME_FIELD: self.username,
                    "password": self.password,
                },
                content_type="application/json",
            )
            user = User.objects.get(username=self.username)
            assert user.last_login is not None
            assert timezone.now() >= user.last_login


@pytest.mark.django_db
class TestTokenRefreshView(APIViewTestCase):
    view_name = "jwt:token_refresh"

    @pytest.fixture(autouse=True)
    def setUp(self):
        super().setUp()
        self.username = "test_user"
        self.password = "test_password"

        self.user = User.objects.create_user(
            username=self.username,
            password=self.password,
        )

    def test_fields_missing(self):
        res = self.view_post(data={}, content_type="application/json")
        assert res.status_code == 400
        assert "refresh" in res.data

    def test_it_should_return_401_if_token_invalid(self):
        token = RefreshToken()
        del token["exp"]

        res = self.view_post(
            data={"refresh": str(token)}, content_type="application/json"
        )
        assert res.status_code == 401
        assert res.data["code"] == "token_not_valid"

        token.set_exp(lifetime=-timedelta(seconds=1))

        res = self.view_post(
            data={"refresh": str(token)}, content_type="application/json"
        )
        assert res.status_code == 401
        assert res.data["code"] == "token_not_valid"

    def test_it_should_return_access_token_if_everything_ok(self):
        refresh = RefreshToken()
        refresh["test_claim"] = "arst"

        # View returns 200
        now = aware_utcnow() - api_settings.ACCESS_TOKEN_LIFETIME / 2

        with patch("ninja_jwt.tokens.aware_utcnow") as fake_aware_utcnow:
            fake_aware_utcnow.return_value = now

            res = self.view_post(
                data={"refresh": str(refresh)}, content_type="application/json"
            )

        assert res.status_code == 200

        access = AccessToken(res.data["access"])

        assert refresh["test_claim"] == access["test_claim"]
        assert access["exp"] == datetime_to_epoch(
            now + api_settings.ACCESS_TOKEN_LIFETIME
        )


@pytest.mark.django_db
class TestTokenObtainSlidingView(APIViewTestCase):
    view_name = "jwt:token_obtain_sliding"

    @pytest.fixture(autouse=True)
    def setUp(self):
        super().setUp()
        self.username = "test_user"
        self.password = "test_password"

        self.user = User.objects.create_user(
            username=self.username,
            password=self.password,
        )

    def test_fields_missing(self):
        res = self.view_post(data={}, content_type="application/json")
        assert res.status_code == 400
        assert User.USERNAME_FIELD in res.data
        assert "password" in res.data

        res = self.view_post(
            data={User.USERNAME_FIELD: self.username}, content_type="application/json"
        )
        assert res.status_code == 400
        assert "password" in res.data

        res = self.view_post(
            data={"password": self.password}, content_type="application/json"
        )
        assert res.status_code == 400
        assert User.USERNAME_FIELD in res.data

    def test_credentials_wrong(self):
        res = self.view_post(
            data={
                User.USERNAME_FIELD: self.username,
                "password": "test_user",
            },
            content_type="application/json",
        )
        assert res.status_code == 401
        assert "detail" in res.data

    def test_user_inactive(self):
        self.user.is_active = False
        self.user.save()

        res = self.view_post(
            data={
                User.USERNAME_FIELD: self.username,
                "password": self.password,
            },
            content_type="application/json",
        )
        assert res.status_code == 401
        assert "detail" in res.data

    def test_success(self):
        res = self.view_post(
            data={
                User.USERNAME_FIELD: self.username,
                "password": self.password,
            },
            content_type="application/json",
        )
        assert res.status_code == 200
        assert "token" in res.data

    def test_update_last_login(self, monkeypatch):
        self.view_post(
            data={
                User.USERNAME_FIELD: self.username,
                "password": self.password,
            },
            content_type="application/json",
        )

        # verify last_login is not updated
        user = User.objects.get(username=self.username)
        assert user.last_login is None

        # verify last_login is updated
        with monkeypatch.context() as m:
            m.setattr(api_settings, "UPDATE_LAST_LOGIN", True)
            self.view_post(
                data={
                    User.USERNAME_FIELD: self.username,
                    "password": self.password,
                },
                content_type="application/json",
            )
            user = User.objects.get(username=self.username)
            assert user.last_login is not None
            assert timezone.now() >= user.last_login


@pytest.mark.django_db
class TestTokenRefreshSlidingView(APIViewTestCase):
    view_name = "jwt:token_refresh_sliding"

    @pytest.fixture(autouse=True)
    def setUp(self):
        super().setUp()
        self.username = "test_user"
        self.password = "test_password"

        self.user = User.objects.create_user(
            username=self.username,
            password=self.password,
        )

    def test_fields_missing(self):
        res = self.view_post(data={}, content_type="application/json")
        assert res.status_code == 400
        assert "token" in res.data

    def test_it_should_return_401_if_token_invalid(self):
        token = SlidingToken()
        del token["exp"]

        res = self.view_post(
            data={"token": str(token)}, content_type="application/json"
        )
        assert res.status_code == 401
        assert res.data["code"] == "token_not_valid"

        token.set_exp(lifetime=-timedelta(seconds=1))

        res = self.view_post(
            data={"token": str(token)}, content_type="application/json"
        )
        assert res.status_code == 401
        assert res.data["code"] == "token_not_valid"

    def test_it_should_return_401_if_token_has_no_refresh_exp_claim(self):
        token = SlidingToken()
        del token[api_settings.SLIDING_TOKEN_REFRESH_EXP_CLAIM]

        res = self.view_post(
            data={"token": str(token)}, content_type="application/json"
        )
        assert res.status_code == 401
        assert res.data["code"] == "token_not_valid"

    def test_it_should_return_401_if_token_has_refresh_period_expired(self):
        token = SlidingToken()
        token.set_exp(
            api_settings.SLIDING_TOKEN_REFRESH_EXP_CLAIM, lifetime=-timedelta(seconds=1)
        )

        res = self.view_post(
            data={"token": str(token)}, content_type="application/json"
        )
        assert res.status_code == 401
        assert res.data["code"] == "token_not_valid"

    def test_it_should_update_token_exp_claim_if_everything_ok(self):
        now = aware_utcnow()

        token = SlidingToken()
        exp = now + api_settings.SLIDING_TOKEN_LIFETIME - timedelta(seconds=1)
        token.set_exp(
            from_time=now,
            lifetime=api_settings.SLIDING_TOKEN_LIFETIME - timedelta(seconds=1),
        )

        # View returns 200
        res = self.view_post(
            data={"token": str(token)}, content_type="application/json"
        )
        assert res.status_code == 200

        # Expiration claim has moved into future
        new_token = SlidingToken(res.data["token"])
        new_exp = datetime_from_epoch(new_token["exp"])

        assert exp < new_exp


@pytest.mark.django_db
class TestTokenVerifyView(APIViewTestCase):
    view_name = "jwt:token_verify"

    @pytest.fixture(autouse=True)
    def setUp(self):
        super().setUp()
        self.username = "test_user"
        self.password = "test_password"

        self.user = User.objects.create_user(
            username=self.username,
            password=self.password,
        )

    def test_fields_missing(self):
        res = self.view_post(data={}, content_type="application/json")
        assert res.status_code == 400
        assert "token" in res.data

    def test_it_should_return_401_if_token_invalid(self):
        token = SlidingToken()
        del token["exp"]

        res = self.view_post(
            data={"token": str(token)}, content_type="application/json"
        )
        assert res.status_code == 401
        assert res.data["code"] == "token_not_valid"

        token.set_exp(lifetime=-timedelta(seconds=1))

        res = self.view_post(
            data={"token": str(token)}, content_type="application/json"
        )
        assert res.status_code == 401
        assert res.data["code"] == "token_not_valid"

    def test_it_should_return_200_if_everything_okay(self):
        token = RefreshToken()

        res = self.view_post(
            data={"token": str(token)}, content_type="application/json"
        )
        assert res.status_code == 200
        assert len(res.data) == 0

    def test_it_should_ignore_token_type(self):
        token = RefreshToken()
        token[api_settings.TOKEN_TYPE_CLAIM] = "fake_type"

        res = self.view_post(
            data={"token": str(token)}, content_type="application/json"
        )
        assert res.status_code == 200
        assert len(res.data) == 0


@pytest.mark.django_db
class TestTokenBlacklistView(APIViewTestCase):
    view_name = "jwt:token_blacklist"

    @pytest.fixture(autouse=True)
    def setUp(self):
        super().setUp()
        self.username = "test_user"
        self.password = "test_password"

        self.user = User.objects.create_user(
            username=self.username,
            password=self.password,
        )

    def test_fields_missing(self):
        res = self.view_post(data={}, content_type="application/json")
        assert res.status_code == 400
        assert "refresh" in res.data

    def test_it_should_return_401_if_token_invalid(self):
        token = RefreshToken()
        del token["exp"]

        res = self.view_post(
            data={"refresh": str(token)}, content_type="application/json"
        )
        assert res.status_code == 401
        assert res.data["code"] == "token_not_valid"

        token.set_exp(lifetime=-timedelta(seconds=1))

        res = self.view_post(
            data={"refresh": str(token)}, content_type="application/json"
        )
        assert res.status_code == 401
        assert res.data["code"] == "token_not_valid"

    def test_it_should_return_if_everything_ok(self):
        refresh = RefreshToken()
        refresh["test_claim"] = "arst"

        # View returns 200
        now = aware_utcnow() - api_settings.ACCESS_TOKEN_LIFETIME / 2

        with patch("ninja_jwt.tokens.aware_utcnow") as fake_aware_utcnow:
            fake_aware_utcnow.return_value = now

            res = self.view_post(
                data={"refresh": str(refresh)}, content_type="application/json"
            )

        assert res.status_code == 200

        assert res.data == {}

    def test_it_should_return_401_if_token_is_blacklisted(self):
        refresh = RefreshToken()
        refresh["test_claim"] = "arst"

        # View returns 200
        now = aware_utcnow() - api_settings.ACCESS_TOKEN_LIFETIME / 2

        with patch("ninja_jwt.tokens.aware_utcnow") as fake_aware_utcnow:
            fake_aware_utcnow.return_value = now

            res = self.view_post(
                data={"refresh": str(refresh)}, content_type="application/json"
            )

        assert res.status_code == 200

        self.view_name = "jwt:token_refresh"
        res = self.view_post(
            data={"refresh": str(refresh)}, content_type="application/json"
        )
        # make sure other tests are not affected
        del self.view_name

        assert res.status_code == 401
        assert res.data["code"] == "token_not_valid"



from datetime import datetime, timedelta
from typing import Any, Optional, Tuple, Type, Union
from uuid import uuid4

from django.conf import settings
from django.contrib.auth.models import AbstractBaseUser
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from .exceptions import TokenBackendError, TokenError
from .settings import api_settings
from .token_blacklist.models import BlacklistedToken, OutstandingToken
from .utils import aware_utcnow, datetime_from_epoch, datetime_to_epoch, format_lazy


class Token:
    """
    A class which validates and wraps an existing JWT or can be used to build a
    new JWT.
    """

    token_type: Optional[str] = None
    lifetime: Optional[datetime] = None

    def __init__(self, token: Optional[Any] = None, verify: bool = True) -> None:
        """
        !!!! IMPORTANT !!!! MUST raise a TokenError with a user-facing error
        message if the given token is invalid, expired, or otherwise not safe
        to use.
        """
        if self.token_type is None or self.lifetime is None:
            raise TokenError(_("Cannot create token with no type or lifetime"))

        self.token = token
        self.current_time = aware_utcnow()

        # Set up token
        if token is not None:
            # An encoded token was provided
            token_backend = self.get_token_backend()

            # Decode token
            try:
                self.payload = token_backend.decode(token, verify=verify)
            except TokenBackendError:
                raise TokenError(_("Token is invalid or expired"))

            if verify:
                self.verify()
        else:
            # New token.  Skip all the verification steps.
            self.payload = {api_settings.TOKEN_TYPE_CLAIM: self.token_type}

            # Set "exp" and "iat" claims with default value
            self.set_exp(from_time=self.current_time, lifetime=self.lifetime)
            self.set_iat(at_time=self.current_time)

            # Set "jti" claim
            self.set_jti()

    def __repr__(self):
        return repr(self.payload)

    def __getitem__(self, key: str):
        return self.payload[key]

    def __setitem__(self, key: str, value: Any):
        self.payload[key] = value

    def __delitem__(self, key: str):
        del self.payload[key]

    def __contains__(self, key: str):
        return key in self.payload

    def get(self, key: str, default=None):
        return self.payload.get(key, default)

    def __str__(self) -> str:
        """
        Signs and returns a token as a base64 encoded string.
        """
        return self.get_token_backend().encode(self.payload)

    def verify(self) -> Any:
        """
        Performs additional validation steps which were not performed when this
        token was decoded.  This method is part of the "public" API to indicate
        the intention that it may be overridden in subclasses.
        """
        # According to RFC 7519, the "exp" claim is OPTIONAL
        # (https://tools.ietf.org/html/rfc7519#section-4.1.4).  As a more
        # correct behavior for authorization tokens, we require an "exp"
        # claim.  We don't want any zombie tokens walking around.
        self.check_exp()

        # Ensure token id is present
        if api_settings.JTI_CLAIM not in self.payload:
            raise TokenError(_("Token has no id"))

        self.verify_token_type()

    def verify_token_type(self) -> Any:
        """
        Ensures that the token type claim is present and has the correct value.
        """
        try:
            token_type = self.payload[api_settings.TOKEN_TYPE_CLAIM]
        except KeyError:
            raise TokenError(_("Token has no type"))

        if self.token_type != token_type:
            raise TokenError(_("Token has wrong type"))

    def set_jti(self) -> None:
        """
        Populates the configured jti claim of a token with a string where there
        is a negligible probability that the same string will be chosen at a
        later time.

        See here:
        https://tools.ietf.org/html/rfc7519#section-4.1.7
        """
        self.payload[api_settings.JTI_CLAIM] = uuid4().hex

    def set_exp(
        self,
        claim: str = "exp",
        from_time: Optional[datetime] = None,
        lifetime: Optional[datetime] = None,
    ) -> None:
        """
        Updates the expiration time of a token.

        See here:
        https://tools.ietf.org/html/rfc7519#section-4.1.4
        """
        if from_time is None:
            from_time = self.current_time

        if lifetime is None:
            lifetime = self.lifetime

        self.payload[claim] = datetime_to_epoch(from_time + lifetime)

    def set_iat(self, claim: str = "iat", at_time: Optional[datetime] = None) -> None:
        """
        Updates the time at which the token was issued.

        See here:
        https://tools.ietf.org/html/rfc7519#section-4.1.6
        """
        if at_time is None:
            at_time = self.current_time

        self.payload[claim] = datetime_to_epoch(at_time)

    def check_exp(self, claim: str = "exp", current_time: Optional[datetime] = None):
        """
        Checks whether a timestamp value in the given claim has passed (since
        the given datetime value in `current_time`).  Raises a TokenError with
        a user-facing error message if so.
        """
        if current_time is None:
            current_time = self.current_time

        try:
            claim_value = self.payload[claim]
        except KeyError:
            raise TokenError(format_lazy(_("Token has no '{}' claim"), claim))

        claim_time = datetime_from_epoch(claim_value)
        if claim_time <= current_time:
            raise TokenError(format_lazy(_("Token '{}' claim has expired"), claim))

    @classmethod
    def for_user(cls, user: Type[AbstractBaseUser]) -> Union["Token", Type["Token"]]:
        """
        Returns an authorization token for the given user that will be provided
        after authenticating the user's credentials.
        """
        user_id = getattr(user, api_settings.USER_ID_FIELD)
        if not isinstance(user_id, int):
            user_id = str(user_id)

        token = cls()
        token[api_settings.USER_ID_CLAIM] = user_id

        return token

    _token_backend = None

    def get_token_backend(self):
        if self._token_backend is None:
            self._token_backend = import_string("ninja_jwt.state.token_backend")
        return self._token_backend


class BlacklistMixin:
    """
    If the `ninja_jwt.token_blacklist` app was configured to be
    used, tokens created from `BlacklistMixin` subclasses will insert
    themselves into an outstanding token list and also check for their
    membership in a token blacklist.
    """

    if "ninja_jwt.token_blacklist" in settings.INSTALLED_APPS:

        def verify(self, *args, **kwargs):
            self.check_blacklist()

            super().verify(*args, **kwargs)

        def check_blacklist(self) -> None:
            """
            Checks if this token is present in the token blacklist.  Raises
            `TokenError` if so.
            """
            jti = self.payload[api_settings.JTI_CLAIM]

            if BlacklistedToken.objects.filter(token__jti=jti).exists():
                raise TokenError(_("Token is blacklisted"))

        def blacklist(self) -> BlacklistedToken:
            """
            Ensures this token is included in the outstanding token list and
            adds it to the blacklist.
            """
            jti = self.payload[api_settings.JTI_CLAIM]
            exp = self.payload["exp"]

            # Ensure outstanding token exists with given jti
            token, _ = OutstandingToken.objects.get_or_create(
                jti=jti,
                defaults={
                    "token": str(self),
                    "expires_at": datetime_from_epoch(exp),
                },
            )

            return BlacklistedToken.objects.get_or_create(token=token)

        @classmethod
        def for_user(cls, user: Type["AbstractBaseUser"]) -> Type[Token]:
            """
            Adds this token to the outstanding token list.
            """
            token = super().for_user(user)

            jti = token[api_settings.JTI_CLAIM]
            exp = token["exp"]

            OutstandingToken.objects.create(
                user=user,
                jti=jti,
                token=str(token),
                created_at=token.current_time,
                expires_at=datetime_from_epoch(exp),
            )

            return token


class SlidingToken(BlacklistMixin, Token):
    token_type: str = "sliding"
    lifetime: timedelta = api_settings.SLIDING_TOKEN_LIFETIME

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        if self.token is None:
            # Set sliding refresh expiration claim if new token
            self.set_exp(
                api_settings.SLIDING_TOKEN_REFRESH_EXP_CLAIM,
                from_time=self.current_time,
                lifetime=api_settings.SLIDING_TOKEN_REFRESH_LIFETIME,
            )


class RefreshToken(BlacklistMixin, Token):
    token_type: str = "refresh"
    lifetime: timedelta = api_settings.REFRESH_TOKEN_LIFETIME
    no_copy_claims: Tuple = (
        api_settings.TOKEN_TYPE_CLAIM,
        "exp",
        # Both of these claims are included even though they may be the same.
        # It seems possible that a third party token might have a custom or
        # namespaced JTI claim as well as a default "jti" claim.  In that case,
        # we wouldn't want to copy either one.
        api_settings.JTI_CLAIM,
        "jti",
    )

    @property
    def access_token(self) -> "AccessToken":
        """
        Returns an access token created from this refresh token.  Copies all
        claims present in this refresh token to the new access token except
        those claims listed in the `no_copy_claims` attribute.
        """
        access = AccessToken()

        # Use instantiation time of refresh token as relative timestamp for
        # access token "exp" claim.  This ensures that both a refresh and
        # access token expire relative to the same time if they are created as
        # a pair.
        access.set_exp(from_time=self.current_time)

        no_copy = self.no_copy_claims
        for claim, value in self.payload.items():
            if claim in no_copy:
                continue
            access[claim] = value

        return access


class AccessToken(Token):
    token_type: str = "access"
    lifetime: timedelta = api_settings.ACCESS_TOKEN_LIFETIME


class UntypedToken(Token):
    token_type: str = "untyped"
    lifetime: timedelta = timedelta(seconds=0)

    def verify_token_type(self) -> None:
        """
        Untyped tokens do not verify the "token_type" claim.  This is useful
        when performing general validation of a token's signature and other
        properties which do not relate to the token's intended use.
        """
        pass



import contextlib
import json

from django.db import connection
from django.db.migrations.executor import MigrationExecutor
from django.test import Client, TestCase
from django.test.utils import override_settings

from ninja_jwt.compat import reverse


def client_action_wrapper(action):
    def wrapper_method(self, *args, url=None, **kwargs):
        if not url and self.view_name is None:
            raise ValueError("Must give value for `view_name` property")

        reverse_args = kwargs.pop("reverse_args", tuple())
        reverse_kwargs = kwargs.pop("reverse_kwargs", dict())
        query_string = kwargs.pop("query_string", None)

        if self.header:
            kwargs.update(self.header)
            self.header.clear()

        _url = url or reverse(self.view_name, args=reverse_args, kwargs=reverse_kwargs)
        if query_string is not None:
            url = _url + "?{0}".format(query_string)

        response = getattr(self.client, action)(_url, *args, **kwargs)
        try:
            response.data = json.loads(response.content)
        except (Exception,):
            pass
        return response

    return wrapper_method


class APIViewTestCase:
    client_class = Client
    header = dict()

    def setUp(self):
        self.client = self.client_class()

    def authenticate_with_token(self, type, token):
        """
        Authenticates requests with the given token.
        """
        self.header = dict(HTTP_AUTHORIZATION="{} {}".format(type, token))

    view_name = None

    view_post = client_action_wrapper("post")
    view_get = client_action_wrapper("get")


def override_api_settings(**new_settings):
    from django.conf import settings

    old_settings = getattr(settings, "SIMPLE_JWT", {})
    combined_settings = dict(old_settings)
    combined_settings.update(new_settings)
    return override_settings(SIMPLE_JWT=combined_settings)


class MigrationTestCase:
    migrate_from = None
    migrate_to = None

    def setUp(self):
        self.migrate_from = [self.migrate_from]
        self.migrate_to = [self.migrate_to]

        # Reverse to the original migration
        executor = MigrationExecutor(connection)
        # executor.migrate(self.migrate_from)

        old_apps = executor.loader.project_state(self.migrate_from).apps
        self.setUpBeforeMigration(old_apps)

        # Run the migration to test
        executor.loader.build_graph()
        # executor.migrate(self.migrate_to)

        self.apps = executor.loader.project_state(self.migrate_to).apps

    def setUpBeforeMigration(self, apps):
        pass



from typing import Dict, Optional, Type, cast

from django.conf import settings
from django.contrib.auth import authenticate, get_user_model
from django.contrib.auth.models import AbstractUser, update_last_login
from django.utils.translation import gettext_lazy as _
from ninja_schema import ModelSchema, Schema
from pydantic import root_validator

from ninja_jwt.utils import token_error

from . import exceptions
from .settings import api_settings
from .tokens import RefreshToken, SlidingToken, Token, UntypedToken

if api_settings.BLACKLIST_AFTER_ROTATION:
    from .token_blacklist.models import BlacklistedToken

user_name_field = get_user_model().USERNAME_FIELD  # type: ignore


class TokenObtainSerializer(ModelSchema):
    class Config:
        model = get_user_model()
        include = ["password", user_name_field]

    _user: Optional[Type[AbstractUser]] = None

    _default_error_messages = {
        "no_active_account": _("No active account found with the given credentials")
    }

    @root_validator(pre=True)
    def validate_inputs(cls, values: Dict) -> dict:
        if user_name_field not in values and "password" not in values:
            raise exceptions.ValidationError(
                {
                    user_name_field: f"{user_name_field} is required",
                    "password": "password is required",
                }
            )

        if not values.get(user_name_field):
            raise exceptions.ValidationError(
                {user_name_field: f"{user_name_field} is required"}
            )

        if not values.get("password"):
            raise exceptions.ValidationError({"password": "password is required"})
        return values

    @root_validator
    def validate_schema(cls, values: Dict) -> dict:
        authenticate_kwargs = {
            user_name_field: values[user_name_field],
            "password": values["password"],
        }

        cls._user = authenticate(**authenticate_kwargs)

        if not api_settings.USER_AUTHENTICATION_RULE(cls._user):
            raise exceptions.AuthenticationFailed(
                cls._default_error_messages["no_active_account"]
            )

        return values

    def output_schema(self) -> Type[Schema]:
        raise NotImplementedError(
            "Must implement `output_schema` method for `TokenObtainSerializer` subclasses"
        )

    @classmethod
    def get_token(cls, user: Type[AbstractUser]) -> Type[Token]:
        raise NotImplementedError(
            "Must implement `get_token` method for `TokenObtainSerializer` subclasses"
        )


class TokenObtainPairOutput(Schema):
    refresh: str
    access: str
    username: str


class TokenObtainPairSerializer(TokenObtainSerializer):
    @classmethod
    def get_token(cls, user: Type[AbstractUser]) -> Type[Token]:
        return RefreshToken.for_user(user)

    @root_validator
    def validate_schema(cls, values: Dict) -> dict:
        refresh = cls.get_token(cls._user)
        refresh = cast(RefreshToken, refresh)

        values["refresh"] = str(refresh)
        values["access"] = str(refresh.access_token)

        if api_settings.UPDATE_LAST_LOGIN:
            update_last_login(None, cls._user)

        return values

    def output_schema(self):
        return TokenObtainPairOutput(**self.dict(exclude={"password"}))


class TokenObtainSlidingOutput(Schema):
    token: str
    username: str


class TokenObtainSlidingSerializer(TokenObtainSerializer):
    @classmethod
    def get_token(cls, user: Type[AbstractUser]) -> Type[Token]:
        return SlidingToken.for_user(user)

    @root_validator
    def validate_schema(cls, values: Dict) -> dict:
        token = cls.get_token(cls._user)

        values["token"] = str(token)

        if api_settings.UPDATE_LAST_LOGIN and cls._user:
            update_last_login(cls, cls._user)

        return values

    def output_schema(self):
        return TokenObtainSlidingOutput(**self.dict(exclude={"password"}))


class TokenRefreshSchema(Schema):
    refresh: str

    @root_validator
    def validate_schema(cls, values: Dict) -> dict:
        if not values.get("refresh"):
            raise exceptions.ValidationError({"refresh": "token is required"})
        return values


class TokenRefreshSerializer(Schema):
    refresh: str
    access: Optional[str]

    @root_validator
    @token_error
    def validate_schema(cls, values: Dict) -> dict:
        if not values.get("refresh"):
            raise exceptions.ValidationError({"refresh": "refresh token is required"})

        refresh = RefreshToken(values["refresh"])

        data = {"access": str(refresh.access_token)}

        if api_settings.ROTATE_REFRESH_TOKENS:
            if api_settings.BLACKLIST_AFTER_ROTATION:
                try:
                    # Attempt to blacklist the given refresh token
                    refresh.blacklist()
                except AttributeError:
                    # If blacklist app not installed, `blacklist` method will
                    # not be present
                    pass

            refresh.set_jti()
            refresh.set_exp()
            refresh.set_iat()

            data["refresh"] = str(refresh)

        return data


class TokenRefreshSlidingSchema(Schema):
    token: str

    @root_validator
    def validate_schema(cls, values: Dict) -> dict:
        if not values.get("token"):
            raise exceptions.ValidationError({"token": "token is required"})
        return values


class TokenRefreshSlidingSerializer(Schema):
    token: str

    @root_validator
    @token_error
    def validate_schema(cls, values: Dict) -> dict:
        if not values.get("token"):
            raise exceptions.ValidationError({"token": "token is required"})

        token = SlidingToken(values["token"])

        # Check that the timestamp in the "refresh_exp" claim has not
        # passed
        token.check_exp(api_settings.SLIDING_TOKEN_REFRESH_EXP_CLAIM)

        # Update the "exp" and "iat" claims
        token.set_exp()
        token.set_iat()

        return {"token": str(token)}


class TokenVerifySerializer(Schema):
    token: str

    @root_validator
    @token_error
    def validate_schema(cls, values: Dict) -> dict:
        if not values.get("token"):
            raise exceptions.ValidationError({"token": "token is required"})
        token = UntypedToken(values["token"])

        if (
            api_settings.BLACKLIST_AFTER_ROTATION
            and "ninja_jwt.token_blacklist" in settings.INSTALLED_APPS
        ):
            jti = token.get(api_settings.JTI_CLAIM)
            if BlacklistedToken.objects.filter(token__jti=jti).exists():
                raise exceptions.ValidationError("Token is blacklisted")

        return values


class TokenBlacklistSerializer(Schema):
    refresh: str

    @root_validator
    @token_error
    def validate_schema(cls, values: Dict) -> dict:
        if not values.get("refresh"):
            raise exceptions.ValidationError({"refresh": "refresh token is required"})
        refresh = RefreshToken(values["refresh"])
        try:
            refresh.blacklist()
        except AttributeError:
            pass
        return values



def pytest_configure():
    from django.conf import settings

    MIDDLEWARE = (
        "django.middleware.common.CommonMiddleware",
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "django.contrib.messages.middleware.MessageMiddleware",
    )

    settings.configure(
        DEBUG_PROPAGATE_EXCEPTIONS=True,
        DATABASES={
            "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"},
            "other": {"ENGINE": "django.db.backends.sqlite3", "NAME": "other"},
        },
        SITE_ID=1,
        SECRET_KEY="not very secret in tests",
        USE_I18N=True,
        USE_L10N=True,
        STATIC_URL="/static/",
        ROOT_URLCONF="tests.urls",
        TEMPLATES=[
            {
                "BACKEND": "django.template.backends.django.DjangoTemplates",
                "APP_DIRS": True,
            },
        ],
        MIDDLEWARE=MIDDLEWARE,
        MIDDLEWARE_CLASSES=MIDDLEWARE,
        INSTALLED_APPS=(
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.sessions",
            "django.contrib.sites",
            "django.contrib.staticfiles",
            "ninja_extra",
            "ninja_jwt",
            "ninja_jwt.token_blacklist",
            "tests",
        ),
        PASSWORD_HASHERS=("django.contrib.auth.hashers.MD5PasswordHasher",),
        SIMPLE_JWT={
            "BLACKLIST_AFTER_ROTATION": True,
        },
    )

    try:
        import django

        django.setup()
    except AttributeError:
        pass



import warnings
from typing import Any

try:
    from django.urls import reverse, reverse_lazy
except ImportError:
    from django.core.urlresolvers import reverse, reverse_lazy  # NOQA


class RemovedInDjango20Warning(DeprecationWarning):
    pass


class CallableBool:  # pragma: no cover
    """
    An boolean-like object that is also callable for backwards compatibility.
    """

    do_not_call_in_templates = True

    def __init__(self, value) -> None:
        self.value = value

    def __bool__(self) -> bool:
        return self.value

    def __call__(self) -> Any:
        warnings.warn(
            "Using user.is_authenticated() and user.is_anonymous() as a method "
            "is deprecated. Remove the parentheses to use it as an attribute.",
            RemovedInDjango20Warning,
            stacklevel=2,
        )
        return self.value

    def __nonzero__(self):  # Python 2 compatibility
        return self.value

    def __repr__(self) -> str:
        return "CallableBool(%r)" % self.value

    def __eq__(self, other) -> bool:
        return self.value == other

    def __ne__(self, other) -> bool:
        return self.value != other

    def __or__(self, other) -> bool:
        return bool(self.value or other)

    def __hash__(self) -> Any:
        return hash(self.value)


CallableFalse = CallableBool(False)
CallableTrue = CallableBool(True)



from datetime import datetime, timedelta
from unittest.mock import patch

from django.test.utils import override_settings
from django.utils import timezone

from ninja_jwt.utils import (
    aware_utcnow,
    datetime_from_epoch,
    datetime_to_epoch,
    format_lazy,
    make_utc,
)


class TestMakeUtc:
    def test_it_should_return_the_correct_values(self):
        # It should make a naive datetime into an aware, utc datetime if django
        # is configured to use timezones and the datetime doesn't already have
        # a timezone

        # Naive datetime
        dt = datetime(year=1970, month=12, day=1)

        with override_settings(USE_TZ=False):
            dt = make_utc(dt)
            assert timezone.is_naive(dt)

        with override_settings(USE_TZ=True):
            dt = make_utc(dt)
            assert timezone.is_aware(dt)
            assert dt.utcoffset() == timedelta(seconds=0)


class TestAwareUtcnow:
    def test_it_should_return_the_correct_value(self):
        now = datetime.utcnow()

        with patch("ninja_jwt.utils.datetime") as fake_datetime:
            fake_datetime.utcnow.return_value = now

            # Should return aware utcnow if USE_TZ == True
            with override_settings(USE_TZ=True):
                assert timezone.make_aware(now, timezone=timezone.utc) == aware_utcnow()

            # Should return naive utcnow if USE_TZ == False
            with override_settings(USE_TZ=False):
                assert now == aware_utcnow()


class TestDatetimeToEpoch:
    def assertEqual(self, value_1, value_2):
        assert value_1 == value_2

    def test_it_should_return_the_correct_values(self):
        self.assertEqual(datetime_to_epoch(datetime(year=1970, month=1, day=1)), 0)
        self.assertEqual(
            datetime_to_epoch(datetime(year=1970, month=1, day=1, second=1)), 1
        )
        self.assertEqual(
            datetime_to_epoch(datetime(year=2000, month=1, day=1)), 946684800
        )


class TestDatetimeFromEpoch:
    def assertEqual(self, value_1, value_2):
        assert value_1 == value_2

    def test_it_should_return_the_correct_values(self):
        with override_settings(USE_TZ=False):
            assert datetime_from_epoch(0) == datetime(year=1970, month=1, day=1)
            assert datetime_from_epoch(1) == datetime(
                year=1970, month=1, day=1, second=1
            )
            assert datetime_from_epoch(946684800) == datetime(year=2000, month=1, day=1)

        with override_settings(USE_TZ=True):
            self.assertEqual(
                datetime_from_epoch(0), make_utc(datetime(year=1970, month=1, day=1))
            )
            self.assertEqual(
                datetime_from_epoch(1),
                make_utc(datetime(year=1970, month=1, day=1, second=1)),
            )
            self.assertEqual(
                datetime_from_epoch(946684800),
                make_utc(datetime(year=2000, month=1, day=1)),
            )


class TestFormatLazy:
    def test_it_should_work(self):
        obj = format_lazy("{} {}", "arst", "zxcv")

        assert not isinstance(obj, str)
        assert str(obj) == "arst zxcv"



from datetime import timedelta
from typing import Any, List, Optional, Union

from django.conf import settings
from django.test.signals import setting_changed
from ninja_extra.lazy import LazyStrImport
from ninja_schema import Schema
from pydantic import AnyUrl, Field, root_validator


class SimpleJWTUserDefinedSettings:
    def __init__(self, data: dict) -> None:
        self.__dict__ = data


SimpleJWT_SETTINGS_DEFAULTS = dict(
    USER_AUTHENTICATION_RULE="ninja_jwt.authentication.default_user_authentication_rule",
    AUTH_TOKEN_CLASSES=["ninja_jwt.tokens.AccessToken"],
    TOKEN_USER_CLASS="ninja_jwt.models.TokenUser",
)

USER_SETTINGS = SimpleJWTUserDefinedSettings(
    getattr(settings, "SIMPLE_JWT", SimpleJWT_SETTINGS_DEFAULTS)
)


class SimpleJWTSettings(Schema):
    class Config:
        orm_mode = True
        validate_assignment = True

    ACCESS_TOKEN_LIFETIME: timedelta = Field(timedelta(minutes=5))
    REFRESH_TOKEN_LIFETIME: timedelta = Field(timedelta(days=1))
    ROTATE_REFRESH_TOKENS: bool = Field(False)
    BLACKLIST_AFTER_ROTATION: bool = Field(False)
    UPDATE_LAST_LOGIN: bool = Field(False)
    ALGORITHM: str = Field("HS256")
    SIGNING_KEY: str = Field(settings.SECRET_KEY)
    VERIFYING_KEY: Optional[str] = Field(None)
    AUDIENCE: Optional[str] = Field(None)
    ISSUER: Optional[str] = Field(None)
    JWK_URL: Optional[AnyUrl] = Field(None)
    LEEWAY: Union[int, timedelta] = Field(0)

    # AUTH_HEADER_TYPES: Tuple[str] = Field(('Bearer',))
    # AUTH_HEADER_NAME: str = Field('HTTP_AUTHORIZATION')

    USER_ID_FIELD: str = Field("id")
    USER_ID_CLAIM: str = Field("user_id")

    USER_AUTHENTICATION_RULE: Any = Field(
        "ninja_jwt.authentication.default_user_authentication_rule"
    )
    TOKEN_USER_CLASS: Any = Field("ninja_jwt.models.TokenUser")
    AUTH_TOKEN_CLASSES: List[Any] = Field(["ninja_jwt.tokens.AccessToken"])

    TOKEN_TYPE_CLAIM: str = Field("token_type")
    JTI_CLAIM: str = Field("jti")
    SLIDING_TOKEN_REFRESH_EXP_CLAIM: str = Field("refresh_exp")
    SLIDING_TOKEN_LIFETIME: timedelta = Field(timedelta(minutes=5))
    SLIDING_TOKEN_REFRESH_LIFETIME: timedelta = Field(timedelta(minutes=5))

    @root_validator
    def validate_settings(cls, values):
        for item in SimpleJWT_SETTINGS_DEFAULTS.keys():
            if isinstance(values[item], (tuple, list)) and isinstance(
                values[item][0], str
            ):
                values[item] = [LazyStrImport(str(klass)) for klass in values[item]]
            if isinstance(values[item], str):
                values[item] = LazyStrImport(values[item])
        return values


# convert to lazy object
api_settings = SimpleJWTSettings.from_orm(USER_SETTINGS)


def reload_api_settings(*args: Any, **kwargs: Any) -> None:
    global api_settings

    setting, value = kwargs["setting"], kwargs["value"]

    if setting == "SIMPLE_JWT":
        api_settings = SimpleJWTSettings.from_orm(SimpleJWTUserDefinedSettings(value))


setting_changed.connect(reload_api_settings)



import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="BlacklistedToken",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("blacklisted_at", models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name="OutstandingToken",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("jti", models.UUIDField(unique=True)),
                ("token", models.TextField()),
                ("created_at", models.DateTimeField()),
                ("expires_at", models.DateTimeField()),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "ordering": ("user",),
            },
        ),
        migrations.AddField(
            model_name="blacklistedtoken",
            name="token",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                to="token_blacklist.OutstandingToken",
            ),
        ),
    ]



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("token_blacklist", "0004_auto_20171017_2013"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="outstandingtoken",
            name="jti",
        ),
    ]



from django.conf import settings
from django.db import models


class OutstandingToken(models.Model):
    id = models.BigAutoField(primary_key=True, serialize=False)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, null=True, blank=True
    )

    jti = models.CharField(unique=True, max_length=255)
    token = models.TextField()

    created_at = models.DateTimeField(null=True, blank=True)
    expires_at = models.DateTimeField()

    class Meta:
        # Work around for a bug in Django:
        # https://code.djangoproject.com/ticket/19422
        #
        # Also see corresponding ticket:
        # https://github.com/encode/django-rest-framework/issues/705
        abstract = "ninja_jwt.token_blacklist" not in settings.INSTALLED_APPS
        ordering = ("user",)

    def __str__(self):
        return "Token for {} ({})".format(
            self.user,
            self.jti,
        )


class BlacklistedToken(models.Model):
    id = models.BigAutoField(primary_key=True, serialize=False)
    token = models.OneToOneField(OutstandingToken, on_delete=models.CASCADE)

    blacklisted_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        # Work around for a bug in Django:
        # https://code.djangoproject.com/ticket/19422
        #
        # Also see corresponding ticket:
        # https://github.com/encode/django-rest-framework/issues/705
        abstract = "ninja_jwt.token_blacklist" not in settings.INSTALLED_APPS

    def __str__(self):
        return "Blacklisted token for {}".format(self.token.user)



from datetime import timedelta

import pytest
from django.contrib.auth import get_user_model

from ninja_jwt.compat import reverse
from ninja_jwt.settings import api_settings
from ninja_jwt.tokens import AccessToken

from .utils import APIViewTestCase, override_api_settings

User = get_user_model()


@pytest.mark.django_db
class TestTestView(APIViewTestCase):
    view_name = "jwt:test_view"

    @pytest.fixture(autouse=True)
    def setUp(self):
        super().setUp()
        self.username = "test_user"
        self.password = "test_password"

        self.user = User.objects.create_user(
            username=self.username,
            password=self.password,
        )

    def test_no_authorization(self):
        res = self.view_get()

        assert res.status_code == 401
        assert "Unauthorized" in res.data["detail"]

    def test_wrong_auth_type(self):
        res = self.view_post(
            url=reverse("jwt:token_obtain_sliding"),
            data={
                User.USERNAME_FIELD: self.username,
                "password": self.password,
            },
            content_type="application/json",
        )

        token = res.data["token"]
        self.authenticate_with_token("Wrong", token)

        res = self.view_get()

        assert res.status_code == 401
        assert "Unauthorized" in res.data["detail"]

    def test_expired_token(self, monkeypatch):
        old_lifetime = AccessToken.lifetime
        AccessToken.lifetime = timedelta(seconds=0)
        try:
            res = self.view_post(
                url=reverse("jwt:token_obtain_pair"),
                data={
                    User.USERNAME_FIELD: self.username,
                    "password": self.password,
                },
                content_type="application/json",
            )
        finally:
            AccessToken.lifetime = old_lifetime

        access = res.data["access"]
        self.authenticate_with_token("Bearer", access)

        with monkeypatch.context() as m:
            m.setattr(
                api_settings, "AUTH_TOKEN_CLASSES", ("ninja_jwt.tokens.AccessToken",)
            )
            res = self.view_get()

        assert res.status_code == 401
        assert "token_not_valid" == res.data["code"]

    def test_user_can_get_sliding_token_and_use_it(self, monkeypatch):
        res = self.view_post(
            url=reverse("jwt:token_obtain_sliding"),
            data={
                User.USERNAME_FIELD: self.username,
                "password": self.password,
            },
            content_type="application/json",
        )

        token = res.data["token"]
        self.authenticate_with_token("Bearer", token)

        with monkeypatch.context() as m:
            m.setattr(
                api_settings, "AUTH_TOKEN_CLASSES", ("ninja_jwt.tokens.SlidingToken",)
            )
            res = self.view_get()

        assert res.status_code == 200
        assert res.data["foo"] == "bar"

    def test_user_can_get_access_and_refresh_tokens_and_use_them(self, monkeypatch):
        res = self.view_post(
            url=reverse("jwt:token_obtain_pair"),
            data={
                User.USERNAME_FIELD: self.username,
                "password": self.password,
            },
            content_type="application/json",
        )

        access = res.data["access"]
        refresh = res.data["refresh"]

        self.authenticate_with_token("Bearer", access)

        with monkeypatch.context() as m:
            m.setattr(
                api_settings, "AUTH_TOKEN_CLASSES", ("ninja_jwt.tokens.AccessToken",)
            )
            res = self.view_get()

        assert res.status_code == 200
        assert res.data["foo"] == "bar"

        res = self.view_post(
            url=reverse("jwt:token_refresh"),
            data={"refresh": refresh},
            content_type="application/json",
        )

        access = res.data["access"]

        self.authenticate_with_token("Bearer", access)
        with monkeypatch.context() as m:
            m.setattr(
                api_settings, "AUTH_TOKEN_CLASSES", ("ninja_jwt.tokens.AccessToken",)
            )
            res = self.view_get()

        assert res.status_code == 200
        assert res.data["foo"] == "bar"



from django.core.management.base import BaseCommand

from ninja_jwt.utils import aware_utcnow

from ...models import OutstandingToken


class Command(BaseCommand):
    help = "Flushes any expired tokens in the outstanding token list"

    def handle(self, *args, **kwargs):
        OutstandingToken.objects.filter(expires_at__lte=aware_utcnow()).delete()



import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("token_blacklist", "0006_auto_20171017_2113"),
    ]

    operations = [
        migrations.AlterField(
            model_name="outstandingtoken",
            name="created_at",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="outstandingtoken",
            name="user",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



from uuid import UUID

from django.db import migrations


def populate_jti_hex(apps, schema_editor):
    OutstandingToken = apps.get_model("token_blacklist", "OutstandingToken")

    db_alias = schema_editor.connection.alias
    for token in OutstandingToken.objects.using(db_alias).all():
        token.jti_hex = token.jti.hex
        token.save()


def reverse_populate_jti_hex(apps, schema_editor):  # pragma: no cover
    OutstandingToken = apps.get_model("token_blacklist", "OutstandingToken")

    db_alias = schema_editor.connection.alias
    for token in OutstandingToken.objects.using(db_alias).all():
        token.jti = UUID(hex=token.jti_hex)
        token.save()


class Migration(migrations.Migration):

    dependencies = [
        ("token_blacklist", "0002_outstandingtoken_jti_hex"),
    ]

    operations = [
        migrations.RunPython(populate_jti_hex, reverse_populate_jti_hex),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("token_blacklist", "0007_auto_20171017_2214"),
    ]

    operations = [
        migrations.AlterField(
            model_name="blacklistedtoken",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="outstandingtoken",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
    ]



from datetime import datetime, timedelta
from unittest.mock import patch

import pytest
from django.contrib.auth import get_user_model
from django.test import TestCase
from jose import jwt

from ninja_jwt.exceptions import TokenError
from ninja_jwt.settings import api_settings
from ninja_jwt.state import token_backend
from ninja_jwt.tokens import (
    AccessToken,
    RefreshToken,
    SlidingToken,
    Token,
    UntypedToken,
)
from ninja_jwt.utils import aware_utcnow, datetime_to_epoch, make_utc

from .utils import override_api_settings

User = get_user_model()


class MyToken(Token):
    token_type = "test"
    lifetime = timedelta(days=1)


class TestToken:
    @pytest.fixture(autouse=True)
    def setUp(self):
        self.token = MyToken()

    def test_init_no_token_type_or_lifetime(self):
        class MyTestToken(Token):
            pass

        with pytest.raises(TokenError):
            MyTestToken()

        MyTestToken.token_type = "test"

        with pytest.raises(TokenError):
            MyTestToken()

        del MyTestToken.token_type
        MyTestToken.lifetime = timedelta(days=1)

        with pytest.raises(TokenError):
            MyTestToken()

        MyTestToken.token_type = "test"
        MyTestToken()

    def test_init_no_token_given(self):
        now = make_utc(datetime(year=2000, month=1, day=1))

        with patch("ninja_jwt.tokens.aware_utcnow") as fake_aware_utcnow:
            fake_aware_utcnow.return_value = now
            t = MyToken()

        assert t.current_time == now
        assert t.token is None

        assert len(t.payload) == 4
        assert t.payload["exp"] == datetime_to_epoch(now + MyToken.lifetime)
        assert t.payload["iat"] == datetime_to_epoch(now)
        assert "jti" in t.payload
        assert t.payload[api_settings.TOKEN_TYPE_CLAIM] == MyToken.token_type

    def test_init_token_given(self):
        # Test successful instantiation
        original_now = aware_utcnow()

        with patch("ninja_jwt.tokens.aware_utcnow") as fake_aware_utcnow:
            fake_aware_utcnow.return_value = original_now
            good_token = MyToken()

        good_token["some_value"] = "arst"
        encoded_good_token = str(good_token)

        now = aware_utcnow()

        # Create new token from encoded token
        with patch("ninja_jwt.tokens.aware_utcnow") as fake_aware_utcnow:
            fake_aware_utcnow.return_value = now
            # Should raise no exception
            t = MyToken(encoded_good_token)

        # Should have expected properties
        assert t.current_time == now
        assert t.token == encoded_good_token

        assert len(t.payload) == 5
        assert t["some_value"] == "arst"
        assert t["exp"] == datetime_to_epoch(original_now + MyToken.lifetime)
        assert t["iat"] == datetime_to_epoch(original_now)
        assert t[api_settings.TOKEN_TYPE_CLAIM] == MyToken.token_type
        assert "jti" in t.payload

    def test_init_bad_sig_token_given(self):
        # Test backend rejects encoded token (expired or bad signature)
        payload = {"foo": "bar"}
        payload["exp"] = aware_utcnow() + timedelta(days=1)
        token_1 = jwt.encode(payload, api_settings.SIGNING_KEY, algorithm="HS256")
        payload["foo"] = "baz"
        token_2 = jwt.encode(payload, api_settings.SIGNING_KEY, algorithm="HS256")

        token_2_payload = token_2.rsplit(".", 1)[0]
        token_1_sig = token_1.rsplit(".", 1)[-1]
        invalid_token = token_2_payload + "." + token_1_sig

        with pytest.raises(TokenError):
            MyToken(invalid_token)

    def test_init_bad_sig_token_given_no_verify(self):
        # Test backend rejects encoded token (expired or bad signature)
        payload = {"foo": "bar"}
        payload["exp"] = aware_utcnow() + timedelta(days=1)
        token_1 = jwt.encode(payload, api_settings.SIGNING_KEY, algorithm="HS256")
        payload["foo"] = "baz"
        token_2 = jwt.encode(payload, api_settings.SIGNING_KEY, algorithm="HS256")

        token_2_payload = token_2.rsplit(".", 1)[0]
        token_1_sig = token_1.rsplit(".", 1)[-1]
        invalid_token = token_2_payload + "." + token_1_sig

        t = MyToken(invalid_token, verify=False)

        assert t.payload == payload

    def test_init_expired_token_given(self):
        t = MyToken()
        t.set_exp(lifetime=-timedelta(seconds=1))

        with pytest.raises(TokenError):
            MyToken(str(t))

    def test_init_no_type_token_given(self):
        t = MyToken()
        del t[api_settings.TOKEN_TYPE_CLAIM]

        with pytest.raises(TokenError):
            MyToken(str(t))

    def test_init_wrong_type_token_given(self):
        t = MyToken()
        t[api_settings.TOKEN_TYPE_CLAIM] = "wrong_type"

        with pytest.raises(TokenError):
            MyToken(str(t))

    def test_init_no_jti_token_given(self):
        t = MyToken()
        del t["jti"]

        with pytest.raises(TokenError):
            MyToken(str(t))

    def test_str(self):
        token = MyToken()
        token.set_exp(
            from_time=make_utc(datetime(year=2000, month=1, day=1)),
            lifetime=timedelta(seconds=0),
        )

        # Delete all but one claim.  We want our lives to be easy and for there
        # to only be a couple of possible encodings.  We're only testing that a
        # payload is successfully encoded here, not that it has specific
        # content.
        del token[api_settings.TOKEN_TYPE_CLAIM]
        del token["jti"]
        del token["iat"]

        # Should encode the given token
        encoded_token = str(token)

        # Token could be one of two depending on header dict ordering
        assert encoded_token in (
            "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJleHAiOjk0NjY4NDgwMH0.VKoOnMgmETawjDZwxrQaHG0xHdo6xBodFy6FXJzTVxs",
            "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJleHAiOjk0NjY4NDgwMH0.iqxxOHV63sjeqNR1GDxX3LPvMymfVB76sOIDqTbjAgk",
        )

    def test_repr(self):
        assert repr(self.token) == repr(self.token.payload)

    def test_getitem(self):
        assert self.token["exp"] == self.token.payload["exp"]

    def test_setitem(self):
        self.token["test"] = 1234
        assert self.token.payload["test"] == 1234

    def test_delitem(self):
        self.token["test"] = 1234
        assert self.token.payload["test"] == 1234

        del self.token["test"]
        assert "test" not in self.token

    def test_contains(self):
        assert "exp" in self.token

    def test_get(self):
        self.token["test"] = 1234

        assert 1234 == self.token.get("test")
        assert 1234 == self.token.get("test", 2345)

        assert self.token.get("does_not_exist") is None
        assert 1234 == self.token.get("does_not_exist", 1234)

    def test_set_jti(self):
        token = MyToken()
        old_jti = token["jti"]

        token.set_jti()

        assert "jti" in token
        assert old_jti != token["jti"]

    def test_set_exp(self):
        now = make_utc(datetime(year=2000, month=1, day=1))

        token = MyToken()
        token.current_time = now

        # By default, should add 'exp' claim to token using `self.current_time`
        # and the TOKEN_LIFETIME setting
        token.set_exp()
        assert token["exp"] == datetime_to_epoch(now + MyToken.lifetime)

        # Should allow overriding of beginning time, lifetime, and claim name
        token.set_exp(claim="refresh_exp", from_time=now, lifetime=timedelta(days=1))
        assert "refresh_exp" in token
        assert token["refresh_exp"] == datetime_to_epoch(now + timedelta(days=1))

    def test_set_iat(self):
        now = make_utc(datetime(year=2000, month=1, day=1))

        token = MyToken()
        token.current_time = now

        # By default, should add 'iat' claim to token using `self.current_time`
        token.set_iat()
        assert token["iat"] == datetime_to_epoch(now)

        # Should allow overriding of time and claim name
        token.set_iat(claim="refresh_iat", at_time=now + timedelta(days=1))
        assert "refresh_iat" in token
        assert token["refresh_iat"] == datetime_to_epoch(now + timedelta(days=1))

    def test_check_exp(self):
        token = MyToken()

        # Should raise an exception if no claim of given kind
        with pytest.raises(TokenError):
            token.check_exp("non_existent_claim")

        current_time = token.current_time
        lifetime = timedelta(days=1)
        exp = token.current_time + lifetime

        token.set_exp(lifetime=lifetime)

        # By default, checks 'exp' claim against `self.current_time`.  Should
        # raise an exception if claim has expired.
        token.current_time = exp
        with pytest.raises(TokenError):
            token.check_exp()

        token.current_time = exp + timedelta(seconds=1)
        with pytest.raises(TokenError):
            token.check_exp()

        # Otherwise, should raise no exception
        token.current_time = current_time
        token.check_exp()

        # Should allow specification of claim to be examined and timestamp to
        # compare against

        # Default claim
        with pytest.raises(TokenError):
            token.check_exp(current_time=exp)

        token.set_exp("refresh_exp", lifetime=timedelta(days=1))

        # Default timestamp
        token.check_exp("refresh_exp")

        # Given claim and timestamp
        with pytest.raises(TokenError):
            token.check_exp(
                "refresh_exp", current_time=current_time + timedelta(days=1)
            )
        with pytest.raises(TokenError):
            token.check_exp(
                "refresh_exp", current_time=current_time + timedelta(days=2)
            )

    @pytest.mark.django_db
    def test_for_user(self, monkeypatch):
        username = "test_user"
        user = User.objects.create_user(
            username=username,
            password="test_password",
        )

        token = MyToken.for_user(user)

        user_id = getattr(user, api_settings.USER_ID_FIELD)
        if not isinstance(user_id, int):
            user_id = str(user_id)

        assert token[api_settings.USER_ID_CLAIM] == user_id

        # Test with non-int user id
        with monkeypatch.context() as m:
            m.setattr(api_settings, "USER_ID_FIELD", "username")
            token = MyToken.for_user(user)

        assert token[api_settings.USER_ID_CLAIM] == username

    def test_get_token_backend(self):
        token = MyToken()

        assert token.get_token_backend() == token_backend


class TestSlidingToken:
    def test_init(self):
        # Should set sliding refresh claim and token type claim
        token = SlidingToken()

        assert token[api_settings.SLIDING_TOKEN_REFRESH_EXP_CLAIM] == datetime_to_epoch(
            token.current_time + api_settings.SLIDING_TOKEN_REFRESH_LIFETIME
        )

        assert token[api_settings.TOKEN_TYPE_CLAIM] == "sliding"


class TestAccessToken:
    def test_init(self):
        # Should set token type claim
        token = AccessToken()
        assert token[api_settings.TOKEN_TYPE_CLAIM] == "access"


class TestRefreshToken:
    def test_init(self):
        # Should set token type claim
        token = RefreshToken()
        assert token[api_settings.TOKEN_TYPE_CLAIM] == "refresh"

    def test_access_token(self):
        # Should create an access token from a refresh token
        refresh = RefreshToken()
        refresh["test_claim"] = "arst"

        access = refresh.access_token

        assert isinstance(access, AccessToken)
        assert access[api_settings.TOKEN_TYPE_CLAIM] == "access"

        # Should keep all copyable claims from refresh token
        assert refresh["test_claim"] == access["test_claim"]

        # Should not copy certain claims from refresh token
        for claim in RefreshToken.no_copy_claims:
            assert refresh[claim] != access[claim]


class TestUntypedToken:
    def test_it_should_accept_and_verify_any_type_of_token(self):
        access_token = AccessToken()
        refresh_token = RefreshToken()
        sliding_token = SlidingToken()

        for t in (access_token, refresh_token, sliding_token):
            untyped_token = UntypedToken(str(t))

            assert t.payload == untyped_token.payload

    def test_it_should_expire_immediately_if_made_from_scratch(self):
        t = UntypedToken()

        assert t[api_settings.TOKEN_TYPE_CLAIM] == "untyped"

        with pytest.raises(TokenError):
            t.check_exp()



from datetime import datetime, timedelta
from unittest import mock
from unittest.mock import patch

import jwt
import pytest
from jwt import PyJWS, algorithms

from ninja_jwt.backends import TokenBackend
from ninja_jwt.exceptions import TokenBackendError
from ninja_jwt.utils import aware_utcnow, datetime_to_epoch, make_utc

from .keys import PRIVATE_KEY, PRIVATE_KEY_2, PUBLIC_KEY, PUBLIC_KEY_2

SECRET = "not_secret"

AUDIENCE = "openid-client-id"

ISSUER = "https://www.myoidcprovider.com"

JWK_URL = "https://randomstring.auth0.com/.well-known/jwks.json"

LEEWAY = 100


class TestTokenBackend:
    @pytest.fixture(autouse=True)
    def setUp(self):
        self.hmac_token_backend = TokenBackend("HS256", SECRET)
        self.hmac_leeway_token_backend = TokenBackend("HS256", SECRET, leeway=LEEWAY)
        self.rsa_token_backend = TokenBackend("RS256", PRIVATE_KEY, PUBLIC_KEY)
        self.aud_iss_token_backend = TokenBackend(
            "RS256", PRIVATE_KEY, PUBLIC_KEY, AUDIENCE, ISSUER
        )
        self.payload = {"foo": "bar"}

    def test_init(self):
        # Should reject unknown algorithms
        with pytest.raises(TokenBackendError):
            TokenBackend("oienarst oieanrsto i", "not_secret")

        TokenBackend("HS256", "not_secret")

    @patch.object(algorithms, "has_crypto", new=False)
    def test_init_fails_for_rs_algorithms_when_crypto_not_installed(self):
        with pytest.raises(
            TokenBackendError,
            match=r"You must have cryptography installed to use RS256.",
        ):
            TokenBackend("RS256", "not_secret")
        with pytest.raises(
            TokenBackendError,
            match=r"You must have cryptography installed to use RS384.",
        ):
            TokenBackend("RS384", "not_secret")
        with pytest.raises(
            TokenBackendError,
            match=r"You must have cryptography installed to use RS512.",
        ):
            TokenBackend("RS512", "not_secret")

    def test_encode_hmac(self):
        # Should return a JSON web token for the given payload
        payload = {"exp": make_utc(datetime(year=2000, month=1, day=1))}

        hmac_token = self.hmac_token_backend.encode(payload)

        # Token could be one of two depending on header dict ordering
        assert hmac_token in (
            "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJleHAiOjk0NjY4NDgwMH0.NHpdD2X8ub4SE_MZLBedWa57FCpntGaN_r6f8kNKdUs",
            "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJleHAiOjk0NjY4NDgwMH0.jvxQgXCSDToR8uKoRJcMT-LmMJJn2-NM76nfSR2FOgs",
        )

    def test_encode_rsa(self):
        # Should return a JSON web token for the given payload
        payload = {"exp": make_utc(datetime(year=2000, month=1, day=1))}

        rsa_token = self.rsa_token_backend.encode(payload)

        # Token could be one of two depending on header dict ordering
        assert rsa_token in (
            "eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiJ9.eyJleHAiOjk0NjY4NDgwMH0.cuE6ocmdxVHXVrGufzn-_ZbXDV475TPPb5jtSacvJsnR3s3tLIm9yR7MF4vGcCAUGJn2hU_JgSOhs9sxntPPVjdwvC3-sxAwfUQ5AgUCEAF5XC7wTvGhmvufzhAgEG_DNsactCh79P8xRnc0iugtlGNyFp_YK582-ZrlfQEp-7C0L9BNG_JCS2J9DsGR7ojO2xGFkFfzezKRkrVTJMPLwpl0JAiZ0iqbQE-Tex7redCrGI388_mgh52GLsNlSIvW2gQMqCVMYndMuYx32Pd5tuToZmLUQ2PJ9RyAZ4fOMApTzoshItg4lGqtnt9CDYzypHULJZohJIPcxFVZZfHxhw",
            "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9.eyJleHAiOjk0NjY4NDgwMH0.pzHTOaVvKJMMkSqksGh-NdeEvQy8Thre3hBM3smUW5Sohtg77KnHpaUYjq30DyRmYQRmPSjEVprh1Yvic_-OeAXPW8WVsF-r4YdJuxWUpuZbIPwJ9E-cMfTZkDkOl18z1zOdlsLtsP2kXyAlptyy9QQsM7AxoqM6cyXoQ5TI0geWccgoahTy3cBtA6pmjm7H0nfeDGqpqYQBhtaFmRuIWn-_XtdN9C6NVmRCcZwyjH-rP3oEm6wtuKJEN25sVWlZm8YRQ-rj7A7SNqBB5tFK2anM_iv4rmBlIEkmr_f2s_WqMxn2EWUSNeqbqiwabR6CZUyJtKx1cPG0B2PqOTcZsg",
        )

    def test_encode_aud_iss(self):
        # Should return a JSON web token for the given payload
        original_payload = {"exp": make_utc(datetime(year=2000, month=1, day=1))}
        payload = original_payload.copy()

        rsa_token = self.aud_iss_token_backend.encode(payload)

        # Assert that payload has not been mutated by the encode() function
        assert payload == original_payload

        # Token could be one of 12 depending on header dict ordering
        assert rsa_token in (
            "eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiJ9.eyJleHAiOjk0NjY4NDgwMCwiYXVkIjoib3BlbmlkLWNsaWVudC1pZCIsImlzcyI6Imh0dHBzOi8vd3d3Lm15b2lkY3Byb3ZpZGVyLmNvbSJ9.kSz7KyUZgpKaeQHYSQlhsE-UFLG2zhBiJ2MFCIvhstA4lSIKj3U1fdP1OhEDg7X66EquRRIZrby6M7RncqCdsjRwKrEIaL74KgC4s5PDXa_HC6dtpi2GhXqaLz8YxfCPaNGZ_9q9rs4Z4O6WpwBLNmMQrTxNno9p0uT93Z2yKj5hGih8a9C_CSf_rKtsHW9AJShWGoKpR6qQFKVNP1GAwQOQ6IeEvZenq_LSEywnrfiWp4Y5UF7xi42wWx7_YPQtM9_Bp5sB-DbrKg_8t0zSc-OHeVDgH0TKqygGEea09W0QkmJcROkaEbxt2LxJg9OuSdXgudVytV8ewpgNtWNE4g",
            "eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiJ9.eyJleHAiOjk0NjY4NDgwMCwiaXNzIjoiaHR0cHM6Ly93d3cubXlvaWRjcHJvdmlkZXIuY29tIiwiYXVkIjoib3BlbmlkLWNsaWVudC1pZCJ9.l-sJR5VKGKNrEn5W8sfO4tEpPq4oQ-Fm5ttQyqUkF6FRJHmCfS1TZIUSXieDLHmarnb4hdIGLr5y-EAbykiqYaTn8d25oT2_zIPlCYHt0DxxeuOliGad5l3AXbWee0qPoZL7YCV8FaSdv2EjtMDOEiJBG5yTkaqZlRmSkbfqu1_y2DRErv3X5LpfEHuKoum4jv5YpoCS6wAWDaWJ9cXMPQaGc4gXg4cuSQxb_EjiQ3QYyztHhG37gOu1J-r_rudaiiIk_VZQdYNfCcirp8isS0z2dcNij_0bELp_oOaipsF7uwzc6WfNGR7xP50X1a_K6EBZzVs0eXFxvl9b3C_d8A",
            "eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiJ9.eyJhdWQiOiJvcGVuaWQtY2xpZW50LWlkIiwiZXhwIjo5NDY2ODQ4MDAsImlzcyI6Imh0dHBzOi8vd3d3Lm15b2lkY3Byb3ZpZGVyLmNvbSJ9.aTwQEIxSzhI5fN4ffQMzZ6h61Ur-Gzh_gPkgOyyWvMX890Z18tC2RisEjXeL5xDGKe2XiEAVtuJa9CjXB9eJoCxNN1k05C-ph82cco-0m_TbMbs0d1MFnfC9ESr4JKynP_Klxi8bi0iZMazduT15pH4UhRkEGsnp8rOKtlt_8_8UOGBJAzG34161lM4JbZqrZDit1DvQdGxaC0lmMgosKg3NDMECfkPe3pGLJ5F_su5yhQk0xyKNCjYyE2FNlilWoDV2KkGiCWdsFOgRMAJwHZ-cdgPg8Vyh2WthBNblsbRVfDrGtfPX0VuW5B0RhBhvI5Iut34P9kkxKRFo3NaiEg",
            "eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiJ9.eyJhdWQiOiJvcGVuaWQtY2xpZW50LWlkIiwiaXNzIjoiaHR0cHM6Ly93d3cubXlvaWRjcHJvdmlkZXIuY29tIiwiZXhwIjo5NDY2ODQ4MDB9.w46s7u28LgsnmK6K_5O15q1SFkKeRgkkplFLi5idq1z7qJjXUi45qpXIyQw3W8a0k1fwa22WB_0XC1MTo22OK3Z0aGNCI2ZCBxvZGOAc1WcCUae44a9LckPHp80q0Hs03NvjsuRVLGXRwDVHrYxuGnFxQSEMbZ650-MQkfVFIXVzMOOAn5Yl4ntjigLcw8iPEqJPnDLdFUSnObZjRzK1M6mJf0-125pqcFsCJaa49rjdbTtnN-VuGnKmv9wV1GwevRQPWjx2vinETURVO9IyZCDtdaLJkvL7Z5IpToK5jrZPc1UWAR0VR8WeWfussFoHzJF86LxVxnqIeXnqOhq5SQ",
            "eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiJ9.eyJpc3MiOiJodHRwczovL3d3dy5teW9pZGNwcm92aWRlci5jb20iLCJhdWQiOiJvcGVuaWQtY2xpZW50LWlkIiwiZXhwIjo5NDY2ODQ4MDB9.Np_SJYye14vz0cvALvfYNqZXvXMD_gY6kIaxA458kbeu6veC_Ds45uWgjJFhTSYFAFWd3TB6M7qZbWgiO0ycION2-B9Yfgaf82WzNtPfgDhu51w1cbLnvuOSRvgX69Q6Z2i1SByewKaSDw25BaMv9Ty4DBdoCbG62qELnNKjDSQvuHlz8cRJv2I6xJBeOYgZV-YN8Zmxsles44a57Vvcj-DjVouHj5m4LperIxb9islNUQBPRTbvw1d_tR8O8ny0mQqbsWL7e2J-wfwdduVf1kVCPePkhHMM6GLhPIrSoTgMzZuRYLBJ61yphuDK98zTknNBM-Jtn5cMyBwP9JBJvA",
            "eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiJ9.eyJpc3MiOiJodHRwczovL3d3dy5teW9pZGNwcm92aWRlci5jb20iLCJleHAiOjk0NjY4NDgwMCwiYXVkIjoib3BlbmlkLWNsaWVudC1pZCJ9.KJcWZtEluPrkRbYj2i_QmdWpZqGZt63Q8nO4QAJ4B4418ZfmgB6A54_vUmWd3Xc18DYgReLoNPlaOuRXtR7rzlMk0-ADjV0bsca5NwTNAV2F-gR9Xsr9iFlcPMNAYf4CAs85gg7deMIxlbGTAaoft__58ah2_vdd8o_nem1PdzsPC198AYtcwnIV206qpeCNR8S_ZTU46OaHwCoySVRx9E7tNG13YSCmGvBaEqgQHKH82qLXo0KivFrjGmGP0xePFG1B8HCZl-LH1euXGGCp6S48q-tmepX5GJwvzaZNBam4pfxQ0GIHa7z11e7sEN-196iDPCK8NzDcXFwHOAnyaA",
            "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9.eyJleHAiOjk0NjY4NDgwMCwiYXVkIjoib3BlbmlkLWNsaWVudC1pZCIsImlzcyI6Imh0dHBzOi8vd3d3Lm15b2lkY3Byb3ZpZGVyLmNvbSJ9.MfhVcFN-9Rd0j11CLtxopzREKdyJH1loSpD4ibjP6Aco4-iM5C99B6gliPgOldtuevhneXV2I7NGhmZFULaYhulcLrAgKe3Gj_TK-sHvwb62e14ArugmK_FAhN7UqbX8hU9wP42LaWXqA7ps4kkJSx-sfgHqMzCKewlAZWwyZBoFgWEgoheKZ7ILkSGf0jzBZlS_1R0jFRSrlYD9rI1S4Px-HllS0t32wRCSbzkp6aVMRs46S0ePrN1mK3spsuQXtYhE2913ZC7p2KwyTGfC2FOOeJdRJknh6kI3Z7pTcsjN2jnQN3o8vPEkN3wl7MbAgAsHcUV45pvyxn4SNBmTMQ",
            "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9.eyJleHAiOjk0NjY4NDgwMCwiaXNzIjoiaHR0cHM6Ly93d3cubXlvaWRjcHJvdmlkZXIuY29tIiwiYXVkIjoib3BlbmlkLWNsaWVudC1pZCJ9.3NjgS14SGFyJ6cix2XJZFPlcyeSu4LSduEMUIH0grJuCljbhalyoib5s4JnBaK4slKrQv1WHlhnKB737hX1FF7_EwQM3toCf--DBjrIuq5cYK3rzcn71JDe_op3CvClwpVyxd2vQZtQfP_tWqN6cNTuWF8ZQ0rJGug6Zb-NeE5h68YK_9tXLZC_i5anyjjAVONOc3Nd-TeIUBaLQKQXOddw0gcTcA7vg3uS0gXTEDq-_ZkF-v9bn1ua4_lgRPbuaYvrBFbXSCEdvNORPfPz4zfL3XU09q0gmnmXC9nxjUFVX4BjkP_YiCCO42sqUKY4y7STTB_IkK_04e2wntonVZA",
            "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJodHRwczovL3d3dy5teW9pZGNwcm92aWRlci5jb20iLCJleHAiOjk0NjY4NDgwMCwiYXVkIjoib3BlbmlkLWNsaWVudC1pZCJ9.b4pdohov81oqzLyCIp4y7e4VYz7LSez7bH0t1o0Zwzau1uXPYXcasT9lxxNMEEiZwHIovPLyWQ6XvF0bMWTk9vc4PyIpkLqsLBJPsuQ-wYUOD04fECmqUX_JaINqm2pPbohTcOQwl0xsE1UMIKTFBZDL1hEXGEMdW9lrPcXilhbC1ikyMpzsmVh55Q_wL2GqydssnOOcDQTqEkWoKvELJJhBcE-YuQkUp8jEVhF3VZ4jEZkzCErTlyXcfe1qXZHkWtw2QNo9s_SfLuRy_fACOo8XE9pHBoE7rqiSm-FmISgiLO1Jj3Pqq-abjN4SnAbU7PZWcV3fUoO1eYLGORmAcw",
            "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJodHRwczovL3d3dy5teW9pZGNwcm92aWRlci5jb20iLCJhdWQiOiJvcGVuaWQtY2xpZW50LWlkIiwiZXhwIjo5NDY2ODQ4MDB9.yDGMBeee4hj8yDtEvVtIsS4tnkPjDMQADTkNh74wtb3oYPgQNqMRWKngXiwjvW2FmnsCUue2cFzLgTbpqlDq0QKcBP0i_UwBiXk9m2wLo0WRFtgw2zNHYSsowu26sFoEjKLgpPZzKrPlU4pnxqa8u3yqg8vIcSTlpX8t3uDqNqhUKP6x-w6wb25h67XDmnORiMwhaOZE_Gs9-H6uWnKdguTIlU1Tj4CjUEnZgZN7dJORiDnO_vHiAyL5yvRjhp5YK0Pq-TtCj5kWoJsjQiKc4laIcgofAKoq_b62Psns8MhxzAxwM7i0rbQZXXYB0VKMUho88uHlpbSWCZxu415lWw",
            "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9.eyJhdWQiOiJvcGVuaWQtY2xpZW50LWlkIiwiaXNzIjoiaHR0cHM6Ly93d3cubXlvaWRjcHJvdmlkZXIuY29tIiwiZXhwIjo5NDY2ODQ4MDB9.BHSCjFeXS6B7KFJi1LpQMEd3ib4Bp9FY3WcB-v7dtP3Ay0SxQZz_gxIbi-tYiNCBQIlfKcfq6vELOjE1WJ5zxPDQM8uV0Pjl41hqYBu3qFv4649a-o2Cd-MaSPUSUogPxzTh2Bk4IdM3sG1Zbd_At4DR_DQwWJDuChA8duA5yG2XPkZr0YF1ZJ326O_jEowvCJiZpzOpH9QsLVPbiX49jtWTwqQGhvpKEj3ztTLFo8VHO-p8bhOGEph2F73F6-GB0XqiWk2Dm1yKAunJCMsM4qXooWfaX6gj-WFhPI9kEXNFfGmPal5i1gb17YoeinbdV2kjN42oxast2Iaa3CMldw",
            "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9.eyJhdWQiOiJvcGVuaWQtY2xpZW50LWlkIiwiZXhwIjo5NDY2ODQ4MDAsImlzcyI6Imh0dHBzOi8vd3d3Lm15b2lkY3Byb3ZpZGVyLmNvbSJ9.s6sElpfKL8WHWfbD_Kbwiy_ip4O082V8ElZqwugvDpS-7yQ3FTvQ3WXqtVAJc-fBZe4ZsBnrXUWwZV0Nhoe6iWKjEjTPjonQWbWL_WUJmIC2HVz18AOISnqReV2rcuLSHQ2ckhsyktlE9K1Rfj-Hi6f3HzzzePEgTsL2ZdBH6GrcmJVDFKqLLrkvOHShoPp7rcwuFBr0J_S1oqYac5O0B-u0OVnxBXTwij0ThrTGMgVCp2rn6Hk0NvtF6CE49Eu4XP8Ue-APT8l5_SjqX9GcrjkJp8Gif_oyBheL-zRg_v-cU60X6qY9wVolO8WodVPSnlE02XyYhLVxvfK-w5129A",
        )

    def test_decode_hmac_with_no_expiry(self):
        no_exp_token = jwt.encode(self.payload, SECRET, algorithm="HS256")

        self.hmac_token_backend.decode(no_exp_token)

    def test_decode_hmac_with_no_expiry_no_verify(self):
        no_exp_token = jwt.encode(self.payload, SECRET, algorithm="HS256")

        assert (
            self.hmac_token_backend.decode(no_exp_token, verify=False) == self.payload
        )

    def test_decode_hmac_with_expiry(self):
        self.payload["exp"] = aware_utcnow() - timedelta(seconds=1)

        expired_token = jwt.encode(self.payload, SECRET, algorithm="HS256")

        with pytest.raises(TokenBackendError):
            self.hmac_token_backend.decode(expired_token)

    def test_decode_hmac_with_invalid_sig(self):
        self.payload["exp"] = aware_utcnow() + timedelta(days=1)
        token_1 = jwt.encode(self.payload, SECRET, algorithm="HS256")
        self.payload["foo"] = "baz"
        token_2 = jwt.encode(self.payload, SECRET, algorithm="HS256")

        token_2_payload = token_2.rsplit(".", 1)[0]
        token_1_sig = token_1.rsplit(".", 1)[-1]
        invalid_token = token_2_payload + "." + token_1_sig

        with pytest.raises(TokenBackendError):
            self.hmac_token_backend.decode(invalid_token)

    def test_decode_hmac_with_invalid_sig_no_verify(self):
        self.payload["exp"] = aware_utcnow() + timedelta(days=1)
        token_1 = jwt.encode(self.payload, SECRET, algorithm="HS256")
        self.payload["foo"] = "baz"
        token_2 = jwt.encode(self.payload, SECRET, algorithm="HS256")
        # Payload copied
        self.payload["exp"] = datetime_to_epoch(self.payload["exp"])

        token_2_payload = token_2.rsplit(".", 1)[0]
        token_1_sig = token_1.rsplit(".", 1)[-1]
        invalid_token = token_2_payload + "." + token_1_sig

        assert (
            self.hmac_token_backend.decode(invalid_token, verify=False) == self.payload
        )

    def test_decode_hmac_success(self):
        self.payload["exp"] = aware_utcnow() + timedelta(days=1)
        self.payload["foo"] = "baz"

        token = jwt.encode(self.payload, SECRET, algorithm="HS256")
        # Payload copied
        self.payload["exp"] = datetime_to_epoch(self.payload["exp"])

        assert self.hmac_token_backend.decode(token) == self.payload

    def test_decode_rsa_with_no_expiry(self):
        no_exp_token = jwt.encode(self.payload, PRIVATE_KEY, algorithm="RS256")

        self.rsa_token_backend.decode(no_exp_token)

    def test_decode_rsa_with_no_expiry_no_verify(self):
        no_exp_token = jwt.encode(self.payload, PRIVATE_KEY, algorithm="RS256")

        assert (
            self.hmac_token_backend.decode(no_exp_token, verify=False) == self.payload
        )

    def test_decode_rsa_with_expiry(self):
        self.payload["exp"] = aware_utcnow() - timedelta(seconds=1)

        expired_token = jwt.encode(self.payload, PRIVATE_KEY, algorithm="RS256")

        with pytest.raises(TokenBackendError):
            self.rsa_token_backend.decode(expired_token)

    def test_decode_rsa_with_invalid_sig(self):
        self.payload["exp"] = aware_utcnow() + timedelta(days=1)
        token_1 = jwt.encode(self.payload, PRIVATE_KEY, algorithm="RS256")
        self.payload["foo"] = "baz"
        token_2 = jwt.encode(self.payload, PRIVATE_KEY, algorithm="RS256")

        token_2_payload = token_2.rsplit(".", 1)[0]
        token_1_sig = token_1.rsplit(".", 1)[-1]
        invalid_token = token_2_payload + "." + token_1_sig

        with pytest.raises(TokenBackendError):
            self.rsa_token_backend.decode(invalid_token)

    def test_decode_rsa_with_invalid_sig_no_verify(self):
        self.payload["exp"] = aware_utcnow() + timedelta(days=1)
        token_1 = jwt.encode(self.payload, PRIVATE_KEY, algorithm="RS256")
        self.payload["foo"] = "baz"
        token_2 = jwt.encode(self.payload, PRIVATE_KEY, algorithm="RS256")

        token_2_payload = token_2.rsplit(".", 1)[0]
        token_1_sig = token_1.rsplit(".", 1)[-1]
        invalid_token = token_2_payload + "." + token_1_sig
        # Payload copied
        self.payload["exp"] = datetime_to_epoch(self.payload["exp"])

        assert (
            self.hmac_token_backend.decode(invalid_token, verify=False) == self.payload
        )

    def test_decode_rsa_success(self):
        self.payload["exp"] = aware_utcnow() + timedelta(days=1)
        self.payload["foo"] = "baz"

        token = jwt.encode(self.payload, PRIVATE_KEY, algorithm="RS256")
        # Payload copied
        self.payload["exp"] = datetime_to_epoch(self.payload["exp"])

        assert self.rsa_token_backend.decode(token) == self.payload

    def test_decode_aud_iss_success(self):
        self.payload["exp"] = aware_utcnow() + timedelta(days=1)
        self.payload["foo"] = "baz"
        self.payload["aud"] = AUDIENCE
        self.payload["iss"] = ISSUER

        token = jwt.encode(self.payload, PRIVATE_KEY, algorithm="RS256")
        # Payload copied
        self.payload["exp"] = datetime_to_epoch(self.payload["exp"])

        assert self.aud_iss_token_backend.decode(token) == self.payload

    def test_decode_rsa_aud_iss_jwk_success(self):
        self.payload["exp"] = aware_utcnow() + timedelta(days=1)
        self.payload["foo"] = "baz"
        self.payload["aud"] = AUDIENCE
        self.payload["iss"] = ISSUER

        token = jwt.encode(
            self.payload,
            PRIVATE_KEY_2,
            algorithm="RS256",
            headers={"kid": "230498151c214b788dd97f22b85410a5"},
        )
        # Payload copied
        self.payload["exp"] = datetime_to_epoch(self.payload["exp"])

        mock_jwk_module = mock.MagicMock()
        with patch("ninja_jwt.backends.PyJWKClient") as mock_jwk_module:
            mock_jwk_client = mock.MagicMock()
            mock_signing_key = mock.MagicMock()

            mock_jwk_module.return_value = mock_jwk_client
            mock_jwk_client.get_signing_key_from_jwt.return_value = mock_signing_key
            type(mock_signing_key).key = mock.PropertyMock(return_value=PUBLIC_KEY_2)

            # Note the PRIV,PUB care is intentially the original pairing
            jwk_token_backend = TokenBackend(
                "RS256", PRIVATE_KEY, PUBLIC_KEY, AUDIENCE, ISSUER, JWK_URL
            )

            assert jwk_token_backend.decode(token) == self.payload

    def test_decode_when_algorithm_not_available(self):
        token = jwt.encode(self.payload, PRIVATE_KEY, algorithm="RS256")

        pyjwt_without_rsa = PyJWS()
        pyjwt_without_rsa.unregister_algorithm("RS256")

        def _decode(jwt, key, algorithms, options, audience, issuer, leeway):
            return pyjwt_without_rsa.decode(jwt, key, algorithms, options)

        with patch.object(jwt, "decode", new=_decode):
            with pytest.raises(TokenBackendError, match="Invalid algorithm specified"):
                self.rsa_token_backend.decode(token)

    def test_decode_when_token_algorithm_does_not_match(self):
        token = jwt.encode(self.payload, PRIVATE_KEY, algorithm="RS256")

        with pytest.raises(TokenBackendError, match="Invalid algorithm specified"):
            self.hmac_token_backend.decode(token)

    def test_decode_leeway_hmac_fail(self):
        self.payload["exp"] = datetime_to_epoch(
            aware_utcnow() - timedelta(seconds=LEEWAY * 2)
        )

        expired_token = jwt.encode(self.payload, SECRET, algorithm="HS256")

        with pytest.raises(TokenBackendError):
            self.hmac_leeway_token_backend.decode(expired_token)

    def test_decode_leeway_hmac_success(self):
        self.payload["exp"] = datetime_to_epoch(
            aware_utcnow() - timedelta(seconds=LEEWAY / 2)
        )

        expired_token = jwt.encode(self.payload, SECRET, algorithm="HS256")

        assert self.hmac_leeway_token_backend.decode(expired_token) == self.payload



import fnmatch
import os
from pathlib import Path

from django.db import migrations, models  # noqa F401

parent_dir = Path(__file__).resolve(strict=True).parent


class Migration(migrations.Migration):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dependencies = [("token_blacklist", "0010_fix_migrate_to_bigautofield")]
        _m = sorted(fnmatch.filter(os.listdir(parent_dir), "000*.py"))
        if len(_m) == 9:
            self.dependencies.insert(0, ("token_blacklist", os.path.splitext(_m[8])[0]))

    operations = []



from calendar import timegm
from datetime import datetime
from functools import wraps
from importlib import import_module

from django.conf import settings
from django.utils.functional import lazy
from django.utils.timezone import is_naive, make_aware, utc
from six import string_types

from ninja_jwt import exceptions


def token_error(func):
    @wraps(func)
    def _wrap(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except exceptions.TokenError as tex:
            raise exceptions.InvalidToken(str(tex))

    return _wrap


def import_callable(path_or_callable):
    if hasattr(path_or_callable, "__call__"):
        return path_or_callable
    else:
        assert isinstance(path_or_callable, string_types)
        package, attr = path_or_callable.rsplit(".", 1)
        packages = import_module(package)
        return getattr(packages, attr)


def make_utc(dt):
    if settings.USE_TZ and is_naive(dt):
        return make_aware(dt, timezone=utc)

    return dt


def aware_utcnow():
    return make_utc(datetime.utcnow())


def datetime_to_epoch(dt):
    return timegm(dt.utctimetuple())


def datetime_from_epoch(ts):
    return make_utc(datetime.utcfromtimestamp(ts))


def format_lazy(s, *args, **kwargs):
    return s.format(*args, **kwargs)


format_lazy = lazy(format_lazy, str)



from typing import Any, Type

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser, AnonymousUser
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _
from ninja.security import HttpBearer

from .exceptions import AuthenticationFailed, InvalidToken, TokenError
from .settings import api_settings
from .tokens import Token


class JWTBaseAuthentication:
    def __init__(self) -> None:
        super().__init__()
        self.user_model = get_user_model()

    @classmethod
    def get_validated_token(cls, raw_token) -> Type[Token]:
        """
        Validates an encoded JSON web token and returns a validated token
        wrapper object.
        """
        messages = []
        for AuthToken in api_settings.AUTH_TOKEN_CLASSES:
            try:
                return AuthToken(raw_token)
            except TokenError as e:
                messages.append(
                    {
                        "token_class": AuthToken.__name__,
                        "token_type": AuthToken.token_type,
                        "message": e.args[0],
                    }
                )

        raise InvalidToken(
            {
                "detail": _("Given token not valid for any token type"),
                "messages": messages,
            }
        )

    def get_user(self, validated_token) -> Type[AbstractUser]:
        """
        Attempts to find and return a user using the given validated token.
        """
        try:
            user_id = validated_token[api_settings.USER_ID_CLAIM]
        except KeyError:
            raise InvalidToken(_("Token contained no recognizable user identification"))

        try:
            user = self.user_model.objects.get(**{api_settings.USER_ID_FIELD: user_id})
        except self.user_model.DoesNotExist:
            raise AuthenticationFailed(_("User not found"))

        if not user.is_active:
            raise AuthenticationFailed(_("User is inactive"))

        return user

    def jwt_authenticate(self, request: HttpRequest, token: str) -> Type[AbstractUser]:
        request.user = AnonymousUser()
        validated_token = self.get_validated_token(token)
        user = self.get_user(validated_token)
        request.user = user
        return user


class JWTAuth(JWTBaseAuthentication, HttpBearer):
    def authenticate(self, request: HttpRequest, token: str) -> Any:
        return self.jwt_authenticate(request, token)

    def get_user(self, validated_token: Any) -> Type[AbstractUser]:
        """
        Returns a stateless user object which is backed by the given validated
        token.
        """
        if api_settings.USER_ID_CLAIM not in validated_token:
            # The TokenUser class assumes tokens will have a recognizable user
            # identifier claim.
            raise InvalidToken(_("Token contained no recognizable user identification"))

        return api_settings.TOKEN_USER_CLASS(validated_token)


def default_user_authentication_rule(user) -> bool:
    # Prior to Django 1.10, inactive users could be authenticated with the
    # default `ModelBackend`.  As of Django 1.10, the `ModelBackend`
    # prevents inactive users from authenticating.  App designers can still
    # allow inactive users to authenticate by opting for the new
    # `AllowAllUsersModelBackend`.  However, we explicitly prevent inactive
    # users from authenticating to enforce a reasonable policy and provide
    # sensible backwards compatibility with older Django versions.
    return user is not None and user.is_active



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("token_blacklist", "0003_auto_20171017_2007"),
    ]

    operations = [
        migrations.AlterField(
            model_name="outstandingtoken",
            name="jti_hex",
            field=models.CharField(unique=True, max_length=255),
        ),
    ]



from ninja_extra import APIController, route, router
from ninja_extra.permissions import AllowAny
from ninja_schema import Schema

from ninja_jwt import schema


class TokenVerificationController(APIController):
    auto_import = False

    @route.post("/verify", response={200: Schema}, url_name="token_verify")
    def verify_token(self, token: schema.TokenVerifySerializer):
        return {}


class TokenBlackListController(APIController):
    auto_import = False

    @route.post("/blacklist", response={200: Schema}, url_name="token_blacklist")
    def blacklist_token(self, refresh: schema.TokenBlacklistSerializer):
        return {}


class TokenObtainPairController(APIController):
    auto_import = False

    @route.post(
        "/pair", response=schema.TokenObtainPairOutput, url_name="token_obtain_pair"
    )
    def obtain_token(self, user_token: schema.TokenObtainPairSerializer):
        return user_token.output_schema()

    @route.post(
        "/refresh", response=schema.TokenRefreshSerializer, url_name="token_refresh"
    )
    def refresh_token(self, refresh_token: schema.TokenRefreshSchema):
        refresh = schema.TokenRefreshSerializer(**refresh_token.dict())
        return refresh


class TokenObtainSlidingController(TokenObtainPairController):
    auto_import = False

    @route.post(
        "/sliding",
        response=schema.TokenObtainSlidingOutput,
        url_name="token_obtain_sliding",
    )
    def obtain_token(self, user_token: schema.TokenObtainSlidingSerializer):
        return user_token.output_schema()

    @route.post(
        "/sliding/refresh",
        response=schema.TokenRefreshSlidingSerializer,
        url_name="token_refresh_sliding",
    )
    def refresh_token(self, refresh_token: schema.TokenRefreshSlidingSchema):
        refresh = schema.TokenRefreshSlidingSerializer(**refresh_token.dict())
        return refresh


@router("/token", permissions=[AllowAny], tags=["token"])
class SimpleJWTDefaultController(
    TokenVerificationController, TokenObtainPairController
):
    """SimpleJWT Default controller for obtaining and refreshing tokens"""

    auto_import = False


@router("/token", permissions=[AllowAny], tags=["token"])
class SimpleJWTSlidingController(
    TokenVerificationController, TokenObtainSlidingController
):
    """
    SimpleJWT Sliding controller for obtaining and refreshing tokens
    Add 'ninja_jwt.tokens.SlidingToken' in AUTH_TOKEN_CLASSES in Settings
    """

    auto_import = False



# Generated by Django 3.2.3 on 2021-05-27 17:46

from pathlib import Path

from django.db import migrations, models

parent_dir = Path(__file__).resolve(strict=True).parent


class Migration(migrations.Migration):

    dependencies = [
        ("token_blacklist", "0008_migrate_to_bigautofield"),
    ]

    operations = [
        migrations.AlterField(
            model_name="blacklistedtoken",
            name="id",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="outstandingtoken",
            name="id",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("token_blacklist", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="outstandingtoken",
            name="jti_hex",
            field=models.CharField(blank=True, null=True, max_length=255),
        ),
    ]



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class TokenBlacklistConfig(AppConfig):
    name = "ninja_jwt.token_blacklist"
    verbose_name = _("Token Blacklist")
    default_auto_field = "django.db.models.BigAutoField"



from datetime import timedelta

import pytest
from django.contrib.auth import get_user_model

from ninja_jwt import authentication
from ninja_jwt.exceptions import AuthenticationFailed, InvalidToken
from ninja_jwt.models import TokenUser
from ninja_jwt.settings import api_settings
from ninja_jwt.tokens import AccessToken, SlidingToken

User = get_user_model()
AuthToken = api_settings.AUTH_TOKEN_CLASSES[0]


class TestJWTBaseAuthentication:
    @pytest.fixture(autouse=True)
    def setUp(self):
        self.backend = authentication.JWTBaseAuthentication()

    @pytest.mark.django_db
    def test_get_validated_token(self, monkeypatch):
        # Should raise InvalidToken if token not valid
        token = AuthToken()
        token.set_exp(lifetime=-timedelta(days=1))
        with pytest.raises(InvalidToken):
            self.backend.get_validated_token(str(token))

        # Otherwise, should return validated token
        token.set_exp()
        assert self.backend.get_validated_token(str(token)).payload == token.payload

        # Should not accept tokens not included in AUTH_TOKEN_CLASSES
        sliding_token = SlidingToken()
        with monkeypatch.context() as m:
            m.setattr(
                api_settings, "AUTH_TOKEN_CLASSES", ("ninja_jwt.tokens.AccessToken",)
            )
            with pytest.raises(InvalidToken) as e:
                self.backend.get_validated_token(str(sliding_token))

            details = e.value.message
            assert len(details["messages"]) == 1
            assert details["messages"][0] == {
                "token_class": "AccessToken",
                "token_type": "access",
                "message": "Token has wrong type",
            }

        # Should accept tokens included in AUTH_TOKEN_CLASSES
        access_token = AccessToken()
        sliding_token = SlidingToken()
        with monkeypatch.context() as m:
            m.setattr(
                api_settings,
                "AUTH_TOKEN_CLASSES",
                (
                    "ninja_jwt.tokens.AccessToken",
                    "ninja_jwt.tokens.SlidingToken",
                ),
            )
            self.backend.get_validated_token(str(access_token))
            self.backend.get_validated_token(str(sliding_token))

    @pytest.mark.django_db
    def test_get_user(self):
        payload = {"some_other_id": "foo"}

        # Should raise error if no recognizable user identification
        with pytest.raises(InvalidToken):
            self.backend.get_user(payload)

        payload[api_settings.USER_ID_CLAIM] = 42

        # Should raise exception if user not found
        with pytest.raises(AuthenticationFailed):
            self.backend.get_user(payload)

        u = User.objects.create_user(username="markhamill")
        u.is_active = False
        u.save()

        payload[api_settings.USER_ID_CLAIM] = getattr(u, api_settings.USER_ID_FIELD)

        # Should raise exception if user is inactive
        with pytest.raises(AuthenticationFailed):
            self.backend.get_user(payload)

        u.is_active = True
        u.save()

        # Otherwise, should return correct user
        assert self.backend.get_user(payload).id == u.id


class TestJWTAuth:
    @pytest.fixture(autouse=True)
    def setUp(self):
        self.backend = authentication.JWTAuth()

    @pytest.mark.django_db
    def test_get_user(self):
        payload = {"some_other_id": "foo"}

        # Should raise error if no recognizable user identification
        with pytest.raises(InvalidToken):
            self.backend.get_user(payload)

        payload[api_settings.USER_ID_CLAIM] = 42

        # Otherwise, should return a token user object
        user = self.backend.get_user(payload)

        assert isinstance(user, TokenUser)
        assert user.id == 42

    @pytest.mark.django_db
    def test_custom_tokenuser(self, monkeypatch):
        from django.utils.functional import cached_property

        class BobSaget(TokenUser):
            @cached_property
            def username(self):
                return "bsaget"

        with monkeypatch.context() as m:
            m.setattr(api_settings, "TOKEN_USER_CLASS", BobSaget)

            # Should return a token user object
            payload = {api_settings.USER_ID_CLAIM: 42}
            user = self.backend.get_user(payload)

            assert isinstance(user, api_settings.TOKEN_USER_CLASS)
            assert user.id == 42
            assert user.username == "bsaget"



from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from .models import BlacklistedToken, OutstandingToken


class OutstandingTokenAdmin(admin.ModelAdmin):
    list_display = (
        "jti",
        "user",
        "created_at",
        "expires_at",
    )
    search_fields = (
        "user__id",
        "jti",
    )
    ordering = ("user",)

    def get_queryset(self, *args, **kwargs):
        qs = super().get_queryset(*args, **kwargs)

        return qs.select_related("user")

    # Read-only behavior defined below
    actions = None

    def get_readonly_fields(self, *args, **kwargs):
        return [f.name for f in self.model._meta.fields]

    def has_add_permission(self, *args, **kwargs):
        return False

    def has_delete_permission(self, *args, **kwargs):
        return False

    def has_change_permission(self, request, obj=None):
        return request.method in [
            "GET",
            "HEAD",
        ] and super().has_change_permission(  # noqa: W504
            request, obj
        )


admin.site.register(OutstandingToken, OutstandingTokenAdmin)


class BlacklistedTokenAdmin(admin.ModelAdmin):
    list_display = (
        "token_jti",
        "token_user",
        "token_created_at",
        "token_expires_at",
        "blacklisted_at",
    )
    search_fields = (
        "token__user__id",
        "token__jti",
    )
    ordering = ("token__user",)

    def get_queryset(self, *args, **kwargs):
        qs = super().get_queryset(*args, **kwargs)

        return qs.select_related("token__user")

    def token_jti(self, obj):
        return obj.token.jti

    token_jti.short_description = _("jti")
    token_jti.admin_order_field = "token__jti"

    def token_user(self, obj):
        return obj.token.user

    token_user.short_description = _("user")
    token_user.admin_order_field = "token__user"

    def token_created_at(self, obj):
        return obj.token.created_at

    token_created_at.short_description = _("created at")
    token_created_at.admin_order_field = "token__created_at"

    def token_expires_at(self, obj):
        return obj.token.expires_at

    token_expires_at.short_description = _("expires at")
    token_expires_at.admin_order_field = "token__expires_at"


admin.site.register(BlacklistedToken, BlacklistedTokenAdmin)



from django.urls import path
from ninja_extra import NinjaExtraAPI, permissions, router

from ninja_jwt.controller import (
    SimpleJWTDefaultController,
    TokenBlackListController,
    TokenObtainSlidingController,
)

from .views import TestAPIController

sliding_router = router(
    "/token",
    permissions=[permissions.AllowAny],
    tags=["token"],
    controller=TokenObtainSlidingController,
)

blacklist_router = router(
    "/token",
    permissions=[permissions.AllowAny],
    tags=["token"],
    controller=TokenBlackListController,
)

api = NinjaExtraAPI(urls_namespace="jwt")
api.add_controller_router(sliding_router, blacklist_router)
api.register_controllers(SimpleJWTDefaultController, TestAPIController)


urlpatterns = [
    path("api/", api.urls),
]



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['ninja_jwt',
 'ninja_jwt.token_blacklist',
 'ninja_jwt.token_blacklist.management',
 'ninja_jwt.token_blacklist.management.commands',
 'ninja_jwt.token_blacklist.migrations']

package_data = \
{'': ['*'],
 'ninja_jwt': ['locale/*',
               'locale/cs/*',
               'locale/cs/LC_MESSAGES/*',
               'locale/de_CH/*',
               'locale/de_CH/LC_MESSAGES/*',
               'locale/es/*',
               'locale/es/LC_MESSAGES/*',
               'locale/es_AR/*',
               'locale/es_AR/LC_MESSAGES/*',
               'locale/es_CL/*',
               'locale/es_CL/LC_MESSAGES/*',
               'locale/fa_IR/*',
               'locale/fa_IR/LC_MESSAGES/*',
               'locale/fr/*',
               'locale/fr/LC_MESSAGES/*',
               'locale/id_ID/*',
               'locale/id_ID/LC_MESSAGES/*',
               'locale/it_IT/*',
               'locale/it_IT/LC_MESSAGES/*',
               'locale/nl_NL/*',
               'locale/nl_NL/LC_MESSAGES/*',
               'locale/pl_PL/*',
               'locale/pl_PL/LC_MESSAGES/*',
               'locale/pt_BR/*',
               'locale/pt_BR/LC_MESSAGES/*',
               'locale/ru_RU/*',
               'locale/ru_RU/LC_MESSAGES/*',
               'locale/uk_UA/*',
               'locale/uk_UA/LC_MESSAGES/*',
               'locale/zh_Hans/*',
               'locale/zh_Hans/LC_MESSAGES/*']}

install_requires = \
['Django >= 2.1',
 'pyjwt>=2,<3',
 'pyjwt[crypto]',
 'ninja-schema',
 'django-ninja-extra']

extras_require = \
{'test': ['pytest',
          'pytest-cov',
          'pytest-django',
          'black',
          'isort',
          'flake8',
          'mypy',
          'django-stubs',
          'python-jose==3.0.0']}

setup(name='django-ninja-jwt',
      version='5.1.1',
      description='Django Ninja JWT - JSON Web Token for Django-Ninja',
      author='Ezeudoh Tochukwu',
      author_email='tochukwu.ezeudoh@gmail.com',
      url='https://github.com/eadwinCode/django-ninja-jwt',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )



from ninja_extra import APIController, route, router

from ninja_jwt import authentication


@router("/test-view", auth=authentication.JWTAuth())
class TestAPIController(APIController):
    @route.get("/test", url_name="test_view")
    def test(self):
        return {"foo": "bar"}



import pytest
from django.contrib.auth import get_user_model
from ninja_extra import NinjaExtraAPI

from ninja_jwt.controller import SimpleJWTDefaultController
from ninja_jwt.schema import TokenObtainPairSerializer, TokenObtainSerializer

# api = NinjaExtraAPI()
# api.register_controllers(SimpleJWTDefaultController)


@pytest.mark.django_db
class TestSchemas:
    def test_tokenObtainSerializer(self):
        api = NinjaExtraAPI()
        api.register_controllers(SimpleJWTDefaultController)
        # UserModel = get_user_model()
        # UserModel.objects.create_user(username='Emeka', password='password')
        # schema = TokenObtainSerializer(username='Emeka', password='password')
        # print(schema.json())
        # schema = TokenObtainPairSerializer(username='Emeka', password='password')
        # print(schema.json())
        # assert schema.dict()



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("token_blacklist", "0005_remove_outstandingtoken_jti"),
    ]

    operations = [
        migrations.RenameField(
            model_name="outstandingtoken",
            old_name="jti_hex",
            new_name="jti",
        ),
    ]



import pytest
from django.test import TestCase

from ninja_jwt.models import TokenUser
from ninja_jwt.settings import api_settings

AuthToken = api_settings.AUTH_TOKEN_CLASSES[0]


class TestTokenUser:
    @pytest.fixture(autouse=True)
    def setUp(self):
        self.token = AuthToken()
        self.token[api_settings.USER_ID_CLAIM] = 42
        self.token["username"] = "deep-thought"
        self.token["some_other_stuff"] = "arstarst"

        self.user = TokenUser(self.token)

    def test_username(self):
        assert self.user.username == "deep-thought"

    def test_is_active(self):
        assert self.user.is_active

    def test_str(self):
        assert str(self.user) == "TokenUser 42"

    def test_id(self):
        assert self.user.id == 42

    def test_pk(self):
        assert self.user.pk == 42

    def test_is_staff(self):
        payload = {api_settings.USER_ID_CLAIM: 42}
        user = TokenUser(payload)

        assert not user.is_staff

        payload["is_staff"] = True
        user = TokenUser(payload)

        assert user.is_staff

    def test_is_superuser(self):
        payload = {api_settings.USER_ID_CLAIM: 42}
        user = TokenUser(payload)

        assert not user.is_superuser

        payload["is_superuser"] = True
        user = TokenUser(payload)

        assert user.is_superuser

    def test_eq(self):
        user1 = TokenUser({api_settings.USER_ID_CLAIM: 1})
        user2 = TokenUser({api_settings.USER_ID_CLAIM: 2})
        user3 = TokenUser({api_settings.USER_ID_CLAIM: 1})

        assert user1 != user2
        assert user1 == user3

    def test_hash(self):
        assert hash(self.user) == hash(self.user.id)

    def test_not_implemented(self):
        with pytest.raises(NotImplementedError):
            self.user.save()

        with pytest.raises(NotImplementedError):
            self.user.delete()

        with pytest.raises(NotImplementedError):
            self.user.set_password("arst")

        with pytest.raises(NotImplementedError):
            self.user.check_password("arst")

    def test_groups(self):
        assert not self.user.groups.exists()

    def test_user_permissions(self):
        assert not self.user.user_permissions.exists()

    def test_get_group_permissions(self):
        assert len(self.user.get_group_permissions()) == 0

    def test_get_all_permissions(self):
        assert len(self.user.get_all_permissions()) == 0

    def test_has_perm(self):
        assert not self.user.has_perm("test_perm")

    def test_has_perms(self):
        assert not self.user.has_perms(["test_perm"])

    def test_has_module_perms(self):
        assert not self.user.has_module_perms("test_module")

    def test_is_anonymous(self):
        assert not self.user.is_anonymous

    def test_is_authenticated(self):
        assert self.user.is_authenticated

    def test_get_username(self):
        assert self.user.get_username() == "deep-thought"
